"""Functions to update TLE of a satcat ID, and fetch where the file is stored."""
import datetime as dt
import pathlib

import spherapy
from spherapy.util import celestrak, epoch_u, spacetrack


def updateTLEs(sat_id_list: list[int]) -> list[int]:
	"""Fetch most recent TLE for provided list of satcat IDs.

	Fetch most recent TLE for provided list of satcat IDs,
	If credentials are stored, use Spacetrack as TLE source, this will fetch all historical
		TLEs for that satcat ID (from most recent stored TLE up to current)
	If no credentials are stored, use Celestrak as TLE source, this will fetch only the most
		recent TLE

	Args:
		sat_id_list: list of satcat ids to fetch

	Returns:
		list of satcat ids successfully fetched
	"""
	if spacetrack.doCredentialsExist():
		modified_list = spacetrack.updateTLEs(sat_id_list)
	else:
		modified_list = celestrak.updateTLEs(sat_id_list)

	return modified_list

def getTLEFilePaths(sat_id_list:list[int], use_packaged:bool=False) -> list[pathlib.Path]:
	"""Fetch list of paths to TLE files.

	Fetch paths to the file storing the list of TLEs for each provided satcat ID
	If credentials are stored, return path containing all historical TLEs (fetched from Spacetrack)
	If no credentials are stored, return path to .temptle containing only most recent TLE
		(from Celestrak)
	You can force the use of TLE data packaged with spherapy by setting use_packaged=True
		This data will be out of date, and should only be used in examples.

	Args:
		sat_id_list: list of satcat ids to find paths
		use_packaged: use the default TLEs packaged with spherapy, these will be out of date.

	Returns:
		list of paths
	"""
	if use_packaged and spherapy.packaged_TLEs is None:
		raise ValueError('There are no TLEs packaged with spherapy')
	if use_packaged and spherapy.packaged_TLEs is not None:
		try:
			path_list = []
			attempted_sat_id = 0
			for sat_id in sat_id_list:
				attempted_sat_id = sat_id
				path_list.append(spherapy.packaged_TLEs[sat_id])
		except KeyError as e:
			raise KeyError(f'TLE for {attempted_sat_id} is not packaged with spherapy.') from e
		else:
			return path_list
	elif spacetrack.doCredentialsExist():
		return [ spacetrack.getTLEFilePath(sat_id) for sat_id in sat_id_list ]
	return [ celestrak.getTLEFilePath(sat_id) for sat_id in sat_id_list ]

def getStoredEpochLimits(sat_id_list:list[int], use_packaged:bool=False) \
							-> dict[int, None|tuple[dt.datetime, dt.datetime|None]]:
	"""Returns limiting epochs for the stored TLEs for each sat in sat_id_list.

	[description]

	Args:
		sat_id_list (list[int]): [description]
		use_packaged (bool): [description] (default: `False`)

	Returns:
		list[tuple[dt.datetime, dt.datetime]]: [description]
	"""
	terminating_epochs = {}
	if use_packaged and spherapy.packaged_TLEs is not None:
		for sat_id in sat_id_list:
			packaged_path = spherapy.packaged_TLEs[sat_id]
			terminating_epochs[sat_id] = epoch_u.getStoredEpochs(packaged_path)
	elif spacetrack.doCredentialsExist():
		for sat_id in sat_id_list:
			terminating_epochs[sat_id] = spacetrack.getStoredEpochs(sat_id)
	else:
		for sat_id in sat_id_list:
			terminating_epochs[sat_id] = celestrak.getStoredEpochs(sat_id)

	return terminating_epochs



"""Command line scripts for spherapy.

Contains:
 - create_credentials (spherapy-create-credentials)

"""

from spherapy import credentials


def create_credentials():
	"""Helper script to call create credentials function."""
	credentials.createCredentials()

def clear_credentials():
	"""Helper script to call clear credentials function."""
	credentials.clearCredentials()



"""Utility functions for dealing with propagation element set strings."""

from typing import TypedDict


class ElementsLineDict(TypedDict):
	"""TypedDict container for lines of an element set.

	Attributes:
		fields: list of each field of an element set line as strings
		line_str: original whole line string of element set (no newline)
	"""
	fields: list[str]
	line_str: str

def split3LELineIntoFields(line:str) -> ElementsLineDict:
	"""Create an ElementsLineDict from an element set line.

	Args:
		line: string

	Returns:
		ElementsLineDict
	"""
	fields = line.split()
	# insert fields into and store original line in tle dict
	line_dict:ElementsLineDict = {'fields':fields,
							'line_str':line}

	return line_dict

def stringify3LEDict(tle_dict:dict[int, ElementsLineDict]) -> str:
	r"""Turn an element set dict 3LE dict back into a \n delimited string."""
	lines = [line_dict['line_str'] for line_dict in tle_dict.values()]
	return '\n'.join(lines)

def dictify3LEs(lines:list[str]) -> list[dict[int, ElementsLineDict]]:
	"""Turn list of strings into list of dicts storing TLE info.

	dict:
		key: line number (0-2)
		value: dict
			key: 'fields' | 'line_str'
			value: 'list of fields as strings' | original TLE line str
	"""
	if len(lines) == 0:
		raise ValueError("No data")
	if len(lines)%3 != 0:
		# If not obvious what lines relate to eachother, can't make assumption -> abort.
		raise ValueError('Incomplete TLEs present, aborting')

	list_3les = []
	tle:dict[int,ElementsLineDict] = {0:{'fields':[], 'line_str':''},
			1:{'fields':[], 'line_str':''},
			2:{'fields':[], 'line_str':''}}
	for line in lines:
		line_dict = split3LELineIntoFields(line)
		# insert fields into and store original line in tle dict
		tle_line_num = int(line_dict['fields'][0])
		tle[tle_line_num] = line_dict

		if tle_line_num == 2: 		# noqa: PLR2004
			# store parsed tle
			list_3les.append(tle.copy())
			# empty dict
			tle[0] = {'fields':[], 'line_str':''}
			tle[1] = {'fields':[], 'line_str':''}
			tle[2] = {'fields':[], 'line_str':''}

	return list_3les



"""Utility functions for converting different epochs.

Attributes:
	GMST_epoch: [description]
"""

import datetime as dt
import pathlib

import numpy as np

from spherapy.util import elements_u

GMST_epoch = dt.datetime(2000, 1, 1, 12, 0, 0)


def epoch2datetime(epoch_str: str) -> dt.datetime:
	"""Converts a fractional epoch string to a datetime object.

	Args:
	epoch_str:	fractional year epoch

	Returns:
		equivalent datetime
	"""
	if not isinstance(epoch_str, str):
		epoch_str = str(epoch_str)

	year = int(epoch_str[:2])
	if year < 50:  # noqa: PLR2004
		year += 2000
	else:
		year += 1900

	fractional_day_of_year = float(epoch_str[2:])

	base = dt.datetime(year, 1, 1, tzinfo=dt.timezone.utc)
	return base + dt.timedelta(days=fractional_day_of_year) - dt.timedelta(days=1)

def epochEarlierThan(epoch_a:str, epoch_b:str) -> bool:
	"""Check if epoch A is earlier than epoch B.

	Args:
		epoch_a: TLE epoch string A
		epoch_b: TLE epoch string B

	Returns:
		True if epoch A is earlier than epoch B
	"""
	datetime_a = epoch2datetime(epoch_a)
	datetime_b = epoch2datetime(epoch_b)
	return datetime_a < datetime_b

def epochLaterThan(epoch_a:str, epoch_b:str) -> bool:
	"""Check if epoch A is later than epoch B.

	Args:
		epoch_a: TLE epoch string A
		epoch_b: TLE epoch string B

	Returns:
		True if epoch A is later than epoch B
	"""
	datetime_a = epoch2datetime(epoch_a)
	datetime_b = epoch2datetime(epoch_b)
	return datetime_a > datetime_b

def datetime2TLEepoch(date: dt.datetime) -> str:
	"""Converts a datetime to a TLE epoch string.

	Args:
		date: Datetime object

	Returns:
		TLE epoch string, with fractional seconds
	"""
	tzinfo = date.tzinfo
	year_str = str(date.year)[-2:]
	day_str = str(date.timetuple().tm_yday).zfill(3)
	fraction_str = str(
		(date - dt.datetime(date.year, date.month, date.day, tzinfo=tzinfo)).total_seconds()
		/ dt.timedelta(days=1).total_seconds()
	)[1:]

	return year_str + day_str + fraction_str


def datetime2sgp4epoch(date: dt.datetime) -> float:
	"""Converts a datetime to an sgp4 epoch.

	Args:
		date: Datetime object

	Returns:
		SGP4 epoch, with fractional seconds
	"""
	tzinfo = date.tzinfo
	sgp4start = dt.datetime(1949, 12, 31, 0, 0, 0, tzinfo=tzinfo)
	delta = date - sgp4start
	return delta.days + delta.seconds / 86400


def findClosestDatetimeIndices(test_arr:np.ndarray[tuple[int], np.dtype[np.datetime64]],
								source_arr:np.ndarray[tuple[int], np.dtype[np.datetime64]]) \
								-> np.ndarray[tuple[int], np.dtype[np.int64]]:
	"""Find the index of the closest datetime in source arr for each datetime in test_arr.

	Both search_arr and source_arr must be sorted.

	Args:
		test_arr: Mx1 array of datetimes, will find closest time for each element in this array
		source_arr: Nx1: array of datetimes to compare to

	Returns:
		Mx1 array of indices of source_arr
	"""
	# convert to unix timestamps first to save memory
	test_arr_fl = np.vectorize(lambda x:x.timestamp())(test_arr)
	source_arr_fl = np.vectorize(lambda x:x.timestamp())(source_arr)
	idx_arr = np.zeros(test_arr_fl.shape, dtype=np.int64)
	for ii in range(len(test_arr_fl)):
		diff = np.abs(source_arr_fl-test_arr_fl[ii])
		idx_arr[ii] = np.argmin(diff,axis=0)

	return idx_arr

def getStoredEpochs(tle_path:pathlib.Path) -> None|tuple[dt.datetime, dt.datetime|None]:
	"""Return the start and end epoch for tle_path.

	Args:
		tle_path: tle file

	Returns:
		(first epoch datetime, last epoch datetime)
		None if no spacetrack tle stored for sat_id
	"""
	if not tle_path.exists():
		return None

	with tle_path.open('r') as fp:
		lines = fp.readlines()

	first_tle_line_1 = elements_u.split3LELineIntoFields(lines[1])
	last_tle_line_1 = elements_u.split3LELineIntoFields(lines[-2])

	first_epoch_dt = epoch2datetime(first_tle_line_1['fields'][3])
	last_epoch_dt = epoch2datetime(last_tle_line_1['fields'][3])

	return (first_epoch_dt, last_epoch_dt)

def getAllStoredEpochs(tle_path:pathlib.Path) -> None|list[dt.datetime]:
	"""Return the all TLE epochs for tle_path.

	Args:
		tle_path: tle file

	Returns:
		list of datetime object equivalent to the TLE epoch for each TLE
		None if no spacetrack tle stored for sat_id
	"""
	if not tle_path.exists():
		return None

	with tle_path.open('r') as fp:
		lines = fp.readlines()

	tle_epochs_dt = []
	for ii in range(1,len(lines),3):
		tle_line = elements_u.split3LELineIntoFields(lines[ii])
		tle_epochs_dt.append(epoch2datetime(tle_line['fields'][3]))


	return tle_epochs_dt

def getStoredTLEByIdx(tle_path:pathlib.Path,
						idx_list:int|list[int],
						string_only:bool=True) \
							-> list[dict[int, elements_u.ElementsLineDict]]|list[str]|None:
	"""Return the TLE at the Idx within tle_path specifed by idx_list.

	Args:
		tle_path: tle file
		idx_list: list of TLE indices within tle_path
		string_only: if true, return only the raw TLE string,
					if false returns a dict of each field within each line

	Returns:
		list of tle_data
	"""
	if not tle_path.exists():
		return None

	if isinstance(idx_list, int):
		idx_list = [idx_list]

	with tle_path.open('r') as fp:
		lines = fp.readlines()

	tles = []
	for idx in idx_list:
		tle_line_0 = elements_u.split3LELineIntoFields(lines[idx*3+0])
		tle_line_1 = elements_u.split3LELineIntoFields(lines[idx*3+1])
		tle_line_2 = elements_u.split3LELineIntoFields(lines[idx*3+2])

		if string_only:
			tle_data = f"{tle_line_0['line_str']}{tle_line_1['line_str']}{tle_line_2['line_str']}"
		else:
			tle_data = {0:tle_line_0, 1:tle_line_1, 2:tle_line_2} 	#type:ignore

		tles.append(tle_data)

	return tles



from pytest_check import check
import numpy as np
from spherapy.timespan import TimeSpan
import datetime as dt
import astropy.time as astrotime
import astropy.units as u

# Disable logging during testing
import logging
logging.disable(logging.CRITICAL)

## Custom checker
def checkHasattr(a:object, name:str) -> bool:
	__tracebackhide__ = True
	if hasattr(a, name):
		return True
	check.fail(f"check {hasattr({a},{name})} == 4")
	return False

'''
Unit testing for the methods of the TimeSpan class
'''
def test_init():
	'''
	Tests to see that the TimeSpan can be initialised from
	an initial time, a timeperiod, and a time step.
	Makes sure the output is sensible
	'''
	birthday = dt.datetime(1996, 9, 10, tzinfo=dt.timezone.utc)
	t0 = dt.datetime(2000,1,1,12,4,5)

	first_week = TimeSpan(birthday, '1d', '7d')
	first_min = TimeSpan(birthday, '1S', '1M')
	fractional_min = TimeSpan(t0,'1S','1.5M')
	first_year = TimeSpan(birthday, '1d', '365d')
	first_decade = TimeSpan(birthday, '365d','3650d')

	minimum_ts = TimeSpan(t0, '1S', '1S')

	# check lengths are correct
	check.equal(len(first_week), 8)
	check.equal(len(first_min), 61)
	check.equal(len(fractional_min), 91)
	check.equal(len(first_year), 366)
	check.equal(len(first_decade), 11)
	check.equal(len(minimum_ts), 2)

	# check start values are correct
	check.equal(first_week[0], birthday)
	check.equal(first_min[0], birthday)
	check.equal(fractional_min[0], t0.replace(tzinfo=dt.timezone.utc))
	check.equal(first_year[0], birthday)
	check.equal(first_decade[0], birthday)
	check.equal(minimum_ts[0], t0.replace(tzinfo=dt.timezone.utc))

	# check end values are correct
	check.equal(first_week[-1], birthday + dt.timedelta(days=7))
	check.equal(first_min[-1], birthday + dt.timedelta(minutes=1))
	check.equal(fractional_min[-1], t0.replace(tzinfo=dt.timezone.utc) + dt.timedelta(minutes=1.5))
	check.equal(first_year[-1], dt.datetime(1997,9,10, tzinfo=dt.timezone.utc))
	check.equal(first_decade[-1], dt.datetime(2006,9,8, tzinfo=dt.timezone.utc))
	check.equal(minimum_ts[-1], t0.replace(tzinfo=dt.timezone.utc) + dt.timedelta(seconds=1))

def test_getitem():
	t0 = dt.datetime(2013,5,17,11,45,23, tzinfo=dt.timezone.utc)
	ts = TimeSpan(t0, '1S', '1M')

	# NONE
	item = ts[None]
	check.is_true(isinstance(item, np.ndarray))
	check.is_true(np.array_equal(item, ts._timearr))

	# int
	item = ts[1]
	check.is_true(isinstance(item, dt.datetime))
	check.equal(item, t0 + dt.timedelta(seconds=1))

	with check.raises(IndexError):
		item = ts[70]

	# slice
	item = ts[1:4]
	check.is_true(isinstance(item, np.ndarray))
	check.equal(len(item), 3)
	for rdt, edt in zip(item, [t0 + dt.timedelta(seconds=i) for i in range(1,4)]):
		check.equal(rdt, edt)

	# check that slice correctly crop the interval to valid range
	item = ts[30:82]
	check.equal(len(item), 31)

	# list[int]
	item = ts[[3,7,4]]
	check.is_true(isinstance(item, np.ndarray))
	check.equal(len(item), 3)
	for rdt, edt in zip(item, [t0 + dt.timedelta(seconds=i) for i in [3,7,4]]):
		check.equal(rdt, edt)

	with check.raises(IndexError):
		item = ts[[1,70]]

	# np.ndarray
	item = ts[np.array([3,7,4])]
	check.is_true(isinstance(item, np.ndarray))
	check.equal(len(item), 3)
	for rdt, edt in zip(item, [t0 + dt.timedelta(seconds=i) for i in [3,7,4]]):
		check.equal(rdt, edt)

	with check.raises(IndexError):
		item = ts[np.ndarray([1,70])]

	# 2-tuple
	item = ts[(5,20)]
	check.is_true(isinstance(item, np.ndarray))
	check.equal(len(item), 15)
	for rdt, edt in zip(item, [t0 + dt.timedelta(seconds=i) for i in range(5,20)]):
		check.equal(rdt, edt)

	# check that 2-tuple correctly crop the interval to valid range
	item = ts[(55,65)]
	check.equal(len(item), 6)

	# 3-tuple
	item = ts[(5,20,3)]
	check.is_true(isinstance(item, np.ndarray))
	check.equal(len(item), 5)
	for rdt, edt in zip(item, [t0 + dt.timedelta(seconds=i) for i in range(5,20,3)]):
		check.equal(rdt, edt)

	# check that 2-tuple correctly crop the interval to valid range
	item = ts[(44, 62, 8)]
	check.equal(len(item), 3)

	with check.raises(TypeError):
		item = ts[0.0]



def test_len():
	t0 = dt.datetime(2000,1,1,12,4,5)

	first_min = TimeSpan(t0, '1S', '1M')
	non_int_sec_span = TimeSpan(t0, '1.5S', '5S')

	check.equal(len(first_min), len(first_min._timearr))
	check.equal(len(non_int_sec_span), len(non_int_sec_span._timearr))

def test_attributes():
	'''
	Test for existance of required attributes in the timespan class
	'''
	birthday = dt.datetime(1996, 9, 10)
	first_week = TimeSpan(birthday, '1d', '7d')

	# Test attributes exist
	checkHasattr(first_week,'start')
	check.equal(first_week.start, birthday.replace(tzinfo=dt.timezone.utc))
	checkHasattr(first_week,'end')
	check.equal(first_week.end, dt.datetime(1996,9,17,tzinfo=dt.timezone.utc))
	checkHasattr(first_week,'time_step')
	check.equal(first_week.time_step, dt.timedelta(days=1))
	checkHasattr(first_week,'time_period')
	check.equal(first_week.time_period, dt.timedelta(days=7))

def test_nonIntegerSpans():
	birthday = dt.datetime(1996, 9, 10)
	t1 = dt.datetime(2016, 12, 30, 0, 0, 1)

	non_int_day_span = TimeSpan(birthday, '1.5d', '7d')
	non_int_sec_span = TimeSpan(birthday, '1.5S', '5S')

	# check correct number of elements
	check.equal(len(non_int_day_span), 5)
	check.equal(len(non_int_sec_span), 4)

	# Test integer number of steps
	check.equal(len(TimeSpan(t1, '1.75M', '1H')), 35)
	check.equal(TimeSpan(t1, '1.75M', '1H').time_period.total_seconds(), 3570)
	check.equal(TimeSpan(t1, '1.75M', '1H').end, t1.replace(tzinfo=dt.timezone.utc) + dt.timedelta(seconds=3570))

def test_parsing():
	'''
	Tests to see that this module correctly parses tricky input with
	a variety of units
	'''
	t0 = dt.datetime(2016, 12, 30, 0, 0, 1, tzinfo=dt.timezone.utc)

	# Test all desired step and period units
	TimeSpan(t0, '1uS', '10uS')
	TimeSpan(t0, '1mS', '10mS')
	TimeSpan(t0, '1S', '10S')
	TimeSpan(t0, '1M', '10M')
	TimeSpan(t0, '1H', '10H')
	TimeSpan(t0, '1d', '10d')
	TimeSpan(t0, '1d', '70d')
	TimeSpan(t0, '1d', '300d')
	TimeSpan(t0, '1d', '3650d')

	with check.raises(ValueError):
		# Test period smaller than timestep
		TimeSpan(t0, '1d', '1M')
		# Test incorrect step unit
		TimeSpan(t0, '1k', '1d')
		# Test incorrect period unit
		TimeSpan(t0, '1d', '1j')

	# Test non integer value of timestep
	check.equal(len(TimeSpan(t0, '1.5M', '1H')), 41)

	# Test non integer value of period
	check.equal(len(TimeSpan(t0, '1M', '1.5H')), 91)

def test_leaps():

	t0 = dt.datetime(2016, 12, 30, 0, 0, 1, tzinfo=dt.timezone.utc)

	# Check leap second handling
	# Previous leap second Dec 31 2016 23:59:60
	astro_delta = astrotime.Time('2017-01-01T0:0:1') - astrotime.Time('2016-12-30T0:0:1')
	astro_delta.format = 'sec'
	check.not_equal(len(TimeSpan(t0, '1S', '2d'))-1, astro_delta.value)

	# Check leap year handling
	ts = TimeSpan(dt.datetime(2016, 1, 1), '1d', '366d')
	check.equal(len(ts), 367)
	check.equal(ts[-1], dt.datetime(2017,1,1, tzinfo=dt.timezone.utc))

def test_conversions():

	birthday = dt.datetime(1996, 9, 10, tzinfo=dt.timezone.utc)

	first_week = TimeSpan(birthday, '1d', '7d')

	dt_week = first_week.asDatetime()
	astropy_week = first_week.asAstropy()

	check.equal(len(dt_week), len(astropy_week))
	# Assert they are of the correct types (dt/astropy.Time)
	check.is_instance(dt_week[0], type(birthday))
	check.is_instance(astropy_week[0], type(astrotime.Time.now()))
	# Assert the time deltas of each are one day
	check.equal(dt_week[1] - dt_week[0], dt.timedelta(days=1))
	check.is_true(astrotime.TimeDelta(1 * u.day, scale='tai').isclose(astropy_week[1] - astropy_week[0]))
	# Assert the timedelta is consistent throughout the week
	check.equal(dt_week[6] - dt_week[5], dt_week[1] - dt_week[0])
	check.is_true(astrotime.TimeDelta(1 * u.day, scale='tai').isclose(astropy_week[6] - astropy_week[5]))
	# Assert the first day for each is Sept 10, 1996
	check.equal(dt_week[0], dt.datetime(1996, 9, 10, tzinfo=dt.timezone.utc))
	check.is_true(astrotime.Time('1996-09-10').isclose(astropy_week[0]))
	# Assert the last day for each is Sept 17, 1996
	check.equal(dt_week[-1], dt.datetime(1996, 9, 17, tzinfo=dt.timezone.utc))
	check.is_true(astrotime.Time('1996-09-17').isclose(astropy_week[-1]))

def test_closestTo():
	'''
	Tests to see that we can find the closest date/time in the TimeSpan
	to another date
	'''
	t0 = dt.datetime(1998, 6, 5, 3, 4, 25)
	timespan = TimeSpan(t0, '1S', '7d')

	inner_target = dt.datetime(1998, 6, 8, 1, 5, 0)
	pre_target = dt.datetime(1996, 1, 1)
	post_target = dt.datetime(2010, 1, 1)

	half_second = dt.timedelta(microseconds=500000)

	# Test where the target date/time is in the TimeSpan
	res_time, res_index = timespan.getClosest(inner_target)
	check.equal(res_index, (inner_target - t0).total_seconds())
	# Test where the target date/time is not in the TimeSpan
	res_time, res_index = timespan.getClosest(pre_target)
	check.equal(res_index, 0)
	res_time, res_index = timespan.getClosest(post_target)
	check.equal(res_index, len(timespan) - 1)
	# Test where there's a tie: ensure it picks the earlier date
	res_time, res_index = timespan.getClosest(t0 + half_second)
	check.equal(res_index, 0)

def test_timeZones():
	"""
	Tests to see that the initialisation process handles time zonescorrectly
	"""

	# all same time
	# t == t0 == t1 == t2 (all are the same instant in time)
	t = dt.datetime(2008, 6, 5, 2, 4, 25)
	t0 = dt.datetime(2008, 6, 5, 2, 4, 25, tzinfo=dt.timezone.utc)

	tz_offset_0545 = dt.timedelta(hours=5, minutes=45) # Kathmandu
	# time @ 1998-06-05 02:04:25 UTC but expressed in timezone +05:45
	t1 = dt.datetime(2008,6,5,7,49,25, tzinfo=dt.timezone(tz_offset_0545))

	tz_offset_2130 = dt.timedelta(hours=-2, minutes=-30) # Labrador (NDT)
	# time @ 1998-06-05 02:04:25 UTC but expressed in timezone -02:30
	t2 = dt.datetime(2008,6,4,23,34,25, tzinfo=dt.timezone(tz_offset_2130))

	timespan = TimeSpan(t, '1S', '1H')
	timespan_utc = TimeSpan(t0, '1S', '1H')
	timespan_0545 = TimeSpan(t1, '1S', '1H')
	timespan_2130 = TimeSpan(t2, '1S', '1H')

	# check timespan and timespan initialised with UTC time zone have all equal elements
	check.equal(len(timespan_utc), len(timespan))
	for ii in range(len(timespan)):
		check.equal(timespan_utc[ii], timespan[ii])

	# check time offset is applied correctly
	check.equal(timespan_0545[0], t.replace(tzinfo=dt.timezone.utc))
	check.equal(timespan_2130[0], t.replace(tzinfo=dt.timezone.utc))

	# check difference between UTC derived timespan and other timezone derived timespan is always 0
	np.testing.assert_allclose(np.vectorize(lambda x: x.total_seconds())(timespan.asDatetime() - timespan_0545.asDatetime()),
												np.zeros(len(timespan)))
	np.testing.assert_allclose(np.vectorize(lambda x: x.total_seconds())(timespan.asDatetime() - timespan_2130.asDatetime()),
												np.zeros(len(timespan)))


	# all same wall time
	t0 = dt.datetime(2008, 6, 5, 2, 4, 25, tzinfo=dt.timezone.utc)
	tz_offset_0545 = dt.timedelta(hours=5, minutes=45) # Kathmandu +05:45
	t1 = dt.datetime(2008, 6, 5, 2, 4, 25, tzinfo=dt.timezone(tz_offset_0545))
	tz_offset_2130 = dt.timedelta(hours=-2, minutes=-30) # Labrador (NDT) -02:30
	t2 = dt.datetime(2008, 6, 5, 2, 4, 25, tzinfo=dt.timezone(tz_offset_2130))

	timespan_utc = TimeSpan(t0, '1S', '1H')
	timespan_0545 = TimeSpan(t1, '1S', '1H')
	timespan_2130 = TimeSpan(t2, '1S', '1H')

	# check time offset is applied correctly
	check.equal(timespan_0545[0], dt.datetime(2008, 6, 4, 20, 19, 25, tzinfo=dt.timezone.utc))
	check.equal(timespan_2130[0], dt.datetime(2008, 6, 5, 4, 34, 25, tzinfo=dt.timezone.utc))

	# check difference between timespan and timezone aware timespan is always equal to the timezone
	np.testing.assert_allclose(np.vectorize(lambda x: x.total_seconds())(timespan.asDatetime() - timespan_0545.asDatetime()),
												tz_offset_0545.total_seconds()*np.ones(len(timespan)))
	np.testing.assert_allclose(np.vectorize(lambda x: x.total_seconds())(timespan.asDatetime() - timespan_2130.asDatetime()),
												tz_offset_2130.total_seconds()*np.ones(len(timespan)))

def test_UTC():
	t0 = dt.datetime(2008, 6, 5, 2, 4, 25, tzinfo=dt.timezone.utc)

	tz_offset_0545 = dt.timedelta(hours=5, minutes=45) # Kathmandu
	# time @ 1998-06-05 02:04:25 UTC but expressed in timezone +05:45
	t1 = dt.datetime(2008,6,5,7,49,25, tzinfo=dt.timezone(tz_offset_0545))

	timespan_utc = TimeSpan(t0, '1S', '1H')
	timespan_0545 = TimeSpan(t1, '1S', '1H')

	check.equal(timespan_utc[0].tzinfo, dt.timezone.utc)
	check.equal(timespan_utc.start.tzinfo, dt.timezone.utc)
	check.equal(timespan_utc.end.tzinfo, dt.timezone.utc)
	for ii in range(len(timespan_utc)):
		check.equal(timespan_utc[ii].tzinfo,timespan_0545[ii].tzinfo)


def test_secondsSinceStart():
		'''
		Tests that the `seconds_since_start` method works as expected
		'''
		birthday = dt.datetime(1996, 9, 10)
		timespan = TimeSpan(birthday, '1H', '1d')
		seconds_elapsed = timespan.secondsSinceStart()
		check.equal(len(seconds_elapsed), 25)
		check.equal(seconds_elapsed[0], 0)
		np.testing.assert_almost_equal(seconds_elapsed[1], 3600)
		np.testing.assert_almost_equal(seconds_elapsed[-2], 82800)



"""Functions to fetch TLEs from Spacetrack.

Attributes:
	MAX_RETRIES: number of times to try and reach Spacetrack.
"""

import datetime as dt
import logging
import pathlib

from typing import TypedDict
from typing_extensions import NotRequired

import spacetrack as sp

import spherapy
from spherapy.util import elements_u, epoch_u

MAX_RETRIES=3

logger = logging.getLogger(__name__)

class _TLEGetter:
	"""Container around python spacetrack library."""
	def __init__(self, sat_id_list:list[int], user:None|str=None, passwd:None|str=None):
		# initialise the client
		if user is None:
			self.username = spherapy.spacetrack_credentials['user']
		else:
			self.username = user
		if passwd is None:
			self.password = spherapy.spacetrack_credentials['passwd']
		else:
			self.password = passwd
		if self.username is None or self.password is None:
			raise InvalidCredentialsError('No Spacetrack Credentials have been entered')

		self.modified_ids = []
		try:
			self.stc = sp.SpaceTrackClient(self.username, self.password)
			for sat_id in sat_id_list:
				logger.info("Trying to update TLEs for %s", sat_id)
				if not self.checkTLEFileExists(sat_id) or self.getNumPastTLEs(sat_id) == 0:
					res = self.fetchAll(sat_id)
					if res is not None:
						self.modified_ids.append(res)
				else:
					res = self.fetchLatest(sat_id)
					if res is not None:
						self.modified_ids.append(res)

		except sp.AuthenticationError:
			raise InvalidCredentialsError('Username and password are incorrect!') from sp.AuthenticationError 	#noqa: E501

	def getModifiedIDs(self) -> list[int]:
		return self.modified_ids

	def checkTLEFileExists(self, sat_id:int) -> bool:
		return getTLEFilePath(sat_id).exists()

	def fetchAll(self, sat_id:int) -> int|None:
		"""Fetch all TLEs for sat_id.

		Args:
			sat_id: satcat id to fetch

		Returns:
			sat_id if succesfully fetched
			None if couldn't fetch
		"""
		attempt_number = 0
		while attempt_number < MAX_RETRIES:
			try:
				opts = self._calcRequestAllOptions(sat_id)
				logger.info("Requesting TLEs from spacetrack with following options: %s", opts)
				resp_line_iterator = self.stc.tle(**opts)
				resp_lines = list(resp_line_iterator)
				tle_dict_list = elements_u.dictify3LEs(resp_lines)
				self._writeTLEsToNewFile(sat_id, tle_dict_list)
				break
			except TimeoutError:
				attempt_number += 1

		if attempt_number == MAX_RETRIES:
			logger.error("Could not fetch All TLEs for sat %s: failed %s times.",
							sat_id, attempt_number)
			return None

		return sat_id

	def getNumPastTLEs(self, sat_id:int) -> int:
		"""Retrieve number of TLEs stored for sat_id."""
		with getTLEFilePath(sat_id).open('r') as fp:
			lines = fp.readlines()
		return int(len(lines)/3)

	def fetchLatest(self, sat_id:int) -> int|None:
		"""Fetch the latest TLEs for sat_id, and append them to the TLE file.

		Args:
			sat_id: satcat id to fetch

		Returns:
			sat_id if succesfully fetched
			None if couldn't fetch
		"""
		attempt_number = 0
		while attempt_number < MAX_RETRIES:
			try:
				_, days_since_last_epoch = self._findLocalLastEpoch(sat_id)
				if days_since_last_epoch > 0.0:
					opts = self._calcRequestPartialOptions(sat_id)
					logger.info("Requesting TLEs from spacetrack with following options: %s", opts)
					resp_line_iterator = self.stc.tle(**opts)
					resp_lines = list(resp_line_iterator)
					tle_dict_list = elements_u.dictify3LEs(resp_lines)
					self._writeTLEsToFile(sat_id, tle_dict_list)
				break
			except TimeoutError:
				attempt_number += 1
		if attempt_number == MAX_RETRIES:
			logger.error("Could not fetch the latest TLEs for sat %s: failed %s times.",
							sat_id, attempt_number)
			raise TimeoutError(f"Could not fetch the latest TLEs for sat {sat_id}: "
								f"failed {attempt_number} times.")

		return sat_id

	def _findLocalLastEpoch(self, sat_id:int) -> tuple[str, float]:
		"""Find the last TLE epoch in a TLE file."""
		# get penultimate and ultimate epochs
		with getTLEFilePath(sat_id).open('r') as fp:
			lines = fp.readlines()
		while lines[-1] == '':
			lines = lines[:-1]
		last_epoch_line = lines[-2]
		last_tle_epoch = last_epoch_line.split()[3]

		last_epoch_datetime = epoch_u.epoch2datetime(last_tle_epoch)
		delta = dt.datetime.now(tz=dt.timezone.utc) - last_epoch_datetime

		return last_tle_epoch, delta.days

	def _calcRequestAllOptions(self, sat_id:int) -> "_RequestOptions":
		"""Format spacetrack library request options. Request all TLEs."""
		request_options:_RequestOptions = {
				'norad_cat_id': sat_id,
				'orderby': 'epoch asc',
				'limit': 500000,
				'format': '3le',
				'iter_lines': True}

		return request_options

	def _calcRequestPartialOptions(self, sat_id:int) -> "_RequestOptions":
		"""Format spacetrack library request options. Request all TLEs since last epoch."""
		_, days_since_last_epoch = self._findLocalLastEpoch(sat_id)

		request_options:_RequestOptions = {
				'norad_cat_id': sat_id,
				'orderby': 'epoch asc',
				'epoch': f'>now-{days_since_last_epoch+1}',
				'limit': 500000,
				'format': '3le',
				'iter_lines': True}

		return request_options

	def _writeTLEsToFile(self, sat_id:int,
								tle_dict_list:list[dict[int, elements_u.ElementsLineDict]]):
		"""Append TLEs to TLE file."""
		last_epoch, _ = self._findLocalLastEpoch(sat_id)
		first_new_idx = None
		for tle_idx, tle_dict in enumerate(tle_dict_list):
			epoch = tle_dict[1]['fields'][3]
			if epoch_u.epochLaterThan(epoch, last_epoch):
				first_new_idx = tle_idx
				break
		if first_new_idx is not None:
			with getTLEFilePath(sat_id).open('a') as fp:
				for tle_dict in tle_dict_list[first_new_idx:]:
					fp.write('\n')
					fp.write(elements_u.stringify3LEDict(tle_dict))

	def _writeTLEsToNewFile(self, sat_id:int,
									tle_dict_list:list[dict[int, elements_u.ElementsLineDict]]):
		"""New TLE, write entire content to new file."""
		with getTLEFilePath(sat_id).open('w') as fp:
			# don't want to begin or end file with newline
			fp.write(elements_u.stringify3LEDict(tle_dict_list[0]))
			for tle_dict in tle_dict_list[1:]:
				fp.write('\n')
				fp.write(elements_u.stringify3LEDict(tle_dict))

class InvalidCredentialsError(Exception):
	"""Error indicating invalid credentials used to access spacetrack."""
	def __init__(self, message:str): # noqa: D107
		super().__init__(message)

class _RequestOptions(TypedDict):
	norad_cat_id: int
	orderby: str
	epoch: NotRequired[str]
	limit: int
	format: str
	iter_lines: bool

def updateTLEs(sat_id_list:list[int], user:None|str=None, passwd:None|str=None) -> list[int]:
	"""Fetch most recent TLE for satcat IDs from spacetrack.

	Fetch most recent TLEs for provided list of satcat IDs, and append to file.
		If no TLE file yet exists, will download all historical TLEs
	Will try MAX_RETRIES before raising a TimeoutError

	Args:
		sat_id_list: list of satcat ids to fetch
		user: [Optional] overriding spacetrack username to use
		passwd: [Optional] overriding spacetrack password to use

	Returns:
		list:list of satcat ids successfully fetched

	Raises:
		TimeoutError
	"""
	logger.info("Using SPACETRACK to update TLEs")
	g = _TLEGetter(sat_id_list,user=user,passwd=passwd)

	return g.getModifiedIDs()

def getTLEFilePath(sat_id:int) -> pathlib.Path:
	"""Gives path to file where spacetrack TLE is stored.

	Spacetrack TLEs are stored in {satcadID}.tle

	Args:
		sat_id: satcat ID

	Returns:
		path to file
	"""
	return spherapy.tle_dir.joinpath(f'{sat_id}.tle')

def getStoredEpochs(sat_id:int) -> None|tuple[dt.datetime, dt.datetime|None]:
	"""Return the start and end epoch for {sat_id}.tle .

	Args:
		sat_id: satcat id to check

	Returns:
		(first epoch datetime, last epoch datetime)
		None if no spacetrack tle stored for sat_id
	"""
	tle_path = getTLEFilePath(sat_id)

	return epoch_u.getStoredEpochs(tle_path)


def doCredentialsExist() -> bool:
	"""Checks if spacetrack credentials have been loaded into Spherapy.

	Returns:
		bool:
	"""
	user_stored = False
	passwd_stored = False
	if spherapy.spacetrack_credentials['user'] is not None:
		user_stored = True
	if spherapy.spacetrack_credentials['passwd'] is not None:
		passwd_stored = True

	return (user_stored and passwd_stored)



from numpy.typing import NDArray

import pytest
from pytest_check import check
import pathlib
import numpy as np
from spherapy import orbit
from spherapy import timespan
from spherapy import updater
from spherapy.util import orbital_u
import spherapy.util.exceptions as exceptions
import datetime as dt

# Disable logging during testing
import logging
logging.disable(logging.CRITICAL)

	# TODO: test getClosestAttribute
	# TODO: check validity of TLE Gen
	# TODO: check validity of propagated parameters Gen
	# TODO: check validity of Analytical Gen
	# TODO: test non earth orbits
	# TODO: test datetime split across multiple TLE epochs
		# TODO: check listed TLE epochs change at correct point of array
		# TODO: check less than 14 day propagation forward or back is all good

@pytest.fixture(scope='module', autouse=True)
def propagationData(pytestFillPackagedData:None) -> dict: 			#noqa: ARG001
	data = {}
	# Keep number of samples low, no need for lots of samples if only really testing attributes.
	data['t0'] = dt.datetime(2004, 12, 29, 0, 0, 1)
	data['t'] = timespan.TimeSpan(data['t0'], '1S', '2M')
	data['t02'] = dt.datetime(2021, 11, 23, 0, 0, 1)
	data['t1'] = timespan.TimeSpan(data['t02'], '1S', '2M')
	data['pos'] = np.tile(np.linspace(7200,8000,len(data['t']),dtype=np.float64),(3,1)).T

	# From multiple TLEs
	ISS_satcat_id = 25544
	data['ISS_tle_path'] = updater.getTLEFilePaths([ISS_satcat_id], use_packaged=True)[0]

	data['o_tle'] = orbit.Orbit.fromTLE(data['t'], pathlib.Path(data['ISS_tle_path']))
	data['o_tle_astro'] = orbit.Orbit.fromTLE(data['t1'], pathlib.Path(data['ISS_tle_path']), astrobodies=True)

	# From fake TLE
	data['o_ftle'] = orbit.Orbit.fromPropagatedOrbitalParam(data['t'], a=(6378 + 600), ecc=0, inc=45, raan=0, argp=0, mean_nu=0)
	data['o_ftle_astro'] = orbit.Orbit.fromPropagatedOrbitalParam(data['t'], a=(6378 + 600), ecc=0, inc=45, raan=0, argp=0, mean_nu=0, astrobodies=True)

	# Analytic from orbital param
	data['o_analytical'] = orbit.Orbit.fromAnalyticalOrbitalParam(data['t'], a=(6378 + 600), ecc=0, inc=45, raan=0, argp=0, true_nu=0)
	data['o_analytical_astro'] = orbit.Orbit.fromAnalyticalOrbitalParam(data['t'], a=(6378 + 600), ecc=0, inc=45, raan=0, argp=0, true_nu=0, astrobodies=True)

	# From list of position
	data['o_poslist'] = orbit.Orbit.fromListOfPositions(data['t'], data['pos'])
	data['o_poslist_astro'] = orbit.Orbit.fromListOfPositions(data['t'], data['pos'], astrobodies=True)

	# Dummy const pos
	data['o_static'] = orbit.Orbit.fromDummyConstantPosition(data['t'], np.zeros((len(data['t']),3)))
	data['o_static_astro'] = orbit.Orbit.fromDummyConstantPosition(data['t'], np.zeros((len(data['t']),3)), sun_pos=np.ones((len(data['t']),3)), moon_pos=-1*np.ones((len(data['t']),3)))

	return data

def test_TLEGenAttributes(propagationData:dict):
	'''
	Test for existance of required attributes in the orbit class
	'''

	propagationData['o_tle'].name
	propagationData['o_tle'].satcat_id
	propagationData['o_tle'].gen_type
	propagationData['o_tle'].timespan
	propagationData['o_tle'].TLE_epochs
	propagationData['o_tle'].pos
	propagationData['o_tle'].pos_ecef
	propagationData['o_tle'].vel_ecef
	propagationData['o_tle'].vel
	propagationData['o_tle'].lat
	propagationData['o_tle'].lon
	propagationData['o_tle'].sun_pos
	propagationData['o_tle'].moon_pos
	propagationData['o_tle'].alt
	propagationData['o_tle'].eclipse
	propagationData['o_tle'].central_body
	propagationData['o_tle'].period
	propagationData['o_tle'].period_steps
	propagationData['o_tle'].semi_major
	propagationData['o_tle'].ecc
	propagationData['o_tle'].inc
	propagationData['o_tle'].raan
	propagationData['o_tle'].argp
	assert propagationData['o_tle'].timespan is not None
	assert propagationData['o_tle'].satcat_id is not None
	assert propagationData['o_tle'].name is not None
	assert propagationData['o_tle'].gen_type is not None
	assert propagationData['o_tle'].central_body is not None
	assert propagationData['o_tle'].pos is not None
	assert propagationData['o_tle'].vel is not None
	assert propagationData['o_tle'].alt is not None
	assert propagationData['o_tle'].pos_ecef is not None
	assert propagationData['o_tle'].vel_ecef is not None
	assert propagationData['o_tle'].lat is not None
	assert propagationData['o_tle'].lon is not None
	assert propagationData['o_tle'].ecc is not None
	assert propagationData['o_tle'].inc is not None
	assert propagationData['o_tle'].semi_major is not None
	assert propagationData['o_tle'].semi_major is not None
	assert propagationData['o_tle'].raan is not None
	assert propagationData['o_tle'].argp is not None
	assert propagationData['o_tle'].TLE_epochs is not None
	assert propagationData['o_tle'].period is not None
	assert propagationData['o_tle'].period_steps is not None

	propagationData['o_tle_astro'].name
	propagationData['o_tle_astro'].satcat_id
	propagationData['o_tle_astro'].gen_type
	propagationData['o_tle_astro'].timespan
	propagationData['o_tle_astro'].TLE_epochs
	propagationData['o_tle_astro'].pos
	propagationData['o_tle_astro'].pos_ecef
	propagationData['o_tle_astro'].vel_ecef
	propagationData['o_tle_astro'].vel
	propagationData['o_tle_astro'].lat
	propagationData['o_tle_astro'].lon
	propagationData['o_tle_astro'].sun_pos
	propagationData['o_tle_astro'].moon_pos
	propagationData['o_tle_astro'].alt
	propagationData['o_tle_astro'].eclipse
	propagationData['o_tle_astro'].central_body
	propagationData['o_tle_astro'].period
	propagationData['o_tle_astro'].period_steps
	propagationData['o_tle_astro'].semi_major
	propagationData['o_tle_astro'].ecc
	propagationData['o_tle_astro'].inc
	propagationData['o_tle_astro'].raan
	propagationData['o_tle_astro'].argp
	assert propagationData['o_tle_astro'].timespan is not None
	assert propagationData['o_tle_astro'].satcat_id is not None
	assert propagationData['o_tle_astro'].name is not None
	assert propagationData['o_tle_astro'].gen_type is not None
	assert propagationData['o_tle_astro'].central_body is not None
	assert propagationData['o_tle_astro'].pos is not None
	assert propagationData['o_tle_astro'].vel is not None
	assert propagationData['o_tle_astro'].alt is not None
	assert propagationData['o_tle_astro'].pos_ecef is not None
	assert propagationData['o_tle_astro'].vel_ecef is not None
	assert propagationData['o_tle_astro'].lat is not None
	assert propagationData['o_tle_astro'].lon is not None
	assert propagationData['o_tle_astro'].ecc is not None
	assert propagationData['o_tle_astro'].inc is not None
	assert propagationData['o_tle_astro'].semi_major is not None
	assert propagationData['o_tle_astro'].semi_major is not None
	assert propagationData['o_tle_astro'].raan is not None
	assert propagationData['o_tle_astro'].argp is not None
	assert propagationData['o_tle_astro'].TLE_epochs is not None
	assert propagationData['o_tle_astro'].period is not None
	assert propagationData['o_tle_astro'].period_steps is not None
	assert propagationData['o_tle_astro'].sun_pos is not None
	assert propagationData['o_tle_astro'].moon_pos is not None
	assert propagationData['o_tle_astro'].eclipse is not None

def test_FakeTLEGenAttributes(propagationData:dict):
	'''
	Test for existance of required attributes in the orbit class
	'''

	propagationData['o_ftle'].name
	propagationData['o_ftle'].satcat_id
	propagationData['o_ftle'].gen_type
	propagationData['o_ftle'].timespan
	propagationData['o_ftle'].TLE_epochs
	propagationData['o_ftle'].pos
	propagationData['o_ftle'].pos_ecef
	propagationData['o_ftle'].vel_ecef
	propagationData['o_ftle'].vel
	propagationData['o_ftle'].lat
	propagationData['o_ftle'].lon
	propagationData['o_ftle'].sun_pos
	propagationData['o_ftle'].moon_pos
	propagationData['o_ftle'].alt
	propagationData['o_ftle'].eclipse
	propagationData['o_ftle'].central_body
	propagationData['o_ftle'].period
	propagationData['o_ftle'].period_steps
	propagationData['o_ftle'].semi_major
	propagationData['o_ftle'].ecc
	propagationData['o_ftle'].inc
	propagationData['o_ftle'].raan
	propagationData['o_ftle'].argp
	assert propagationData['o_ftle'].name is not None
	assert propagationData['o_ftle'].timespan is not None
	assert propagationData['o_ftle'].gen_type is not None
	assert propagationData['o_ftle'].central_body is not None
	assert propagationData['o_ftle'].pos is not None
	assert propagationData['o_ftle'].vel is not None
	assert propagationData['o_ftle'].alt is not None
	# TODO: include pos_ecef, vel_ecef, lat, lon once these are included in the generation
	assert propagationData['o_ftle'].ecc is not None
	assert propagationData['o_ftle'].inc is not None
	assert propagationData['o_ftle'].semi_major is not None
	assert propagationData['o_ftle'].raan is not None
	assert propagationData['o_ftle'].argp is not None
	assert propagationData['o_ftle'].TLE_epochs is not None
	assert propagationData['o_ftle'].period is not None
	assert propagationData['o_ftle'].period_steps is not None


	propagationData['o_ftle_astro'].name
	propagationData['o_ftle_astro'].satcat_id
	propagationData['o_ftle_astro'].gen_type
	propagationData['o_ftle_astro'].timespan
	propagationData['o_ftle_astro'].TLE_epochs
	propagationData['o_ftle_astro'].pos
	propagationData['o_ftle_astro'].pos_ecef
	propagationData['o_ftle_astro'].vel_ecef
	propagationData['o_ftle_astro'].vel
	propagationData['o_ftle_astro'].lat
	propagationData['o_ftle_astro'].lon
	propagationData['o_ftle_astro'].sun_pos
	propagationData['o_ftle_astro'].moon_pos
	propagationData['o_ftle_astro'].alt
	propagationData['o_ftle_astro'].eclipse
	propagationData['o_ftle_astro'].central_body
	propagationData['o_ftle_astro'].period
	propagationData['o_ftle_astro'].period_steps
	propagationData['o_ftle_astro'].semi_major
	propagationData['o_ftle_astro'].ecc
	propagationData['o_ftle_astro'].inc
	propagationData['o_ftle_astro'].raan
	propagationData['o_ftle_astro'].argp
	assert propagationData['o_ftle_astro'].name is not None
	assert propagationData['o_ftle_astro'].timespan is not None
	assert propagationData['o_ftle_astro'].gen_type is not None
	assert propagationData['o_ftle_astro'].central_body is not None
	assert propagationData['o_ftle_astro'].pos is not None
	assert propagationData['o_ftle_astro'].vel is not None
	assert propagationData['o_ftle_astro'].alt is not None
	# TODO: include pos_ecef, vel_ecef, lat, lon once these are included in the generation
	assert propagationData['o_ftle_astro'].ecc is not None
	assert propagationData['o_ftle_astro'].inc is not None
	assert propagationData['o_ftle_astro'].semi_major is not None
	assert propagationData['o_ftle_astro'].raan is not None
	assert propagationData['o_ftle_astro'].argp is not None
	assert propagationData['o_ftle_astro'].TLE_epochs is not None
	assert propagationData['o_ftle_astro'].period is not None
	assert propagationData['o_ftle_astro'].period_steps is not None
	assert propagationData['o_ftle_astro'].sun_pos is not None
	assert propagationData['o_ftle_astro'].moon_pos is not None
	assert propagationData['o_ftle_astro'].eclipse is not None

def test_AnalyticalGenAttributes(propagationData:dict):
	'''
	Test for existance of required attributes in the orbit class
	'''

	propagationData['o_analytical'].name
	propagationData['o_analytical'].satcat_id
	propagationData['o_analytical'].gen_type
	propagationData['o_analytical'].timespan
	propagationData['o_analytical'].TLE_epochs
	propagationData['o_analytical'].pos
	propagationData['o_analytical'].pos_ecef
	propagationData['o_analytical'].vel_ecef
	propagationData['o_analytical'].vel
	propagationData['o_analytical'].lat
	propagationData['o_analytical'].lon
	propagationData['o_analytical'].sun_pos
	propagationData['o_analytical'].moon_pos
	propagationData['o_analytical'].alt
	propagationData['o_analytical'].eclipse
	propagationData['o_analytical'].central_body
	propagationData['o_analytical'].period
	propagationData['o_analytical'].period_steps
	propagationData['o_analytical'].semi_major
	propagationData['o_analytical'].ecc
	propagationData['o_analytical'].inc
	propagationData['o_analytical'].raan
	propagationData['o_analytical'].argp
	assert propagationData['o_analytical'].name is not None
	assert propagationData['o_analytical'].timespan is not None
	assert propagationData['o_analytical'].gen_type is not None
	assert propagationData['o_analytical'].central_body is not None
	assert propagationData['o_analytical'].pos is not None
	assert propagationData['o_analytical'].vel is not None
	assert propagationData['o_analytical'].alt is not None
	assert propagationData['o_analytical'].ecc is not None
	assert propagationData['o_analytical'].inc is not None
	assert propagationData['o_analytical'].semi_major is not None
	assert propagationData['o_analytical'].raan is not None
	assert propagationData['o_analytical'].argp is not None
	assert propagationData['o_analytical'].period is not None
	assert propagationData['o_analytical'].period_steps is not None

	propagationData['o_analytical_astro'].name
	propagationData['o_analytical_astro'].satcat_id
	propagationData['o_analytical_astro'].gen_type
	propagationData['o_analytical_astro'].timespan
	propagationData['o_analytical_astro'].TLE_epochs
	propagationData['o_analytical_astro'].pos
	propagationData['o_analytical_astro'].pos_ecef
	propagationData['o_analytical_astro'].vel_ecef
	propagationData['o_analytical_astro'].vel
	propagationData['o_analytical_astro'].lat
	propagationData['o_analytical_astro'].lon
	propagationData['o_analytical_astro'].sun_pos
	propagationData['o_analytical_astro'].moon_pos
	propagationData['o_analytical_astro'].alt
	propagationData['o_analytical_astro'].eclipse
	propagationData['o_analytical_astro'].central_body
	propagationData['o_analytical_astro'].period
	propagationData['o_analytical_astro'].period_steps
	propagationData['o_analytical_astro'].semi_major
	propagationData['o_analytical_astro'].ecc
	propagationData['o_analytical_astro'].inc
	propagationData['o_analytical_astro'].raan
	propagationData['o_analytical_astro'].argp
	assert propagationData['o_analytical_astro'].name is not None
	assert propagationData['o_analytical_astro'].timespan is not None
	assert propagationData['o_analytical_astro'].gen_type is not None
	assert propagationData['o_analytical_astro'].central_body is not None
	assert propagationData['o_analytical_astro'].pos is not None
	assert propagationData['o_analytical_astro'].vel is not None
	assert propagationData['o_analytical_astro'].alt is not None
	assert propagationData['o_analytical_astro'].ecc is not None
	assert propagationData['o_analytical_astro'].inc is not None
	assert propagationData['o_analytical_astro'].semi_major is not None
	assert propagationData['o_analytical_astro'].raan is not None
	assert propagationData['o_analytical_astro'].argp is not None
	assert propagationData['o_analytical_astro'].period is not None
	assert propagationData['o_analytical_astro'].period_steps is not None
	assert propagationData['o_analytical_astro'].sun_pos is not None
	assert propagationData['o_analytical_astro'].moon_pos is not None
	assert propagationData['o_analytical_astro'].eclipse is not None

def test_PosListGenAttributes(propagationData:dict):
	'''
	Test for existance of required attributes in the orbit class
	'''

	propagationData['o_poslist'].name
	propagationData['o_poslist'].satcat_id
	propagationData['o_poslist'].gen_type
	propagationData['o_poslist'].timespan
	propagationData['o_poslist'].TLE_epochs
	propagationData['o_poslist'].pos
	propagationData['o_poslist'].pos_ecef
	propagationData['o_poslist'].vel_ecef
	propagationData['o_poslist'].vel
	propagationData['o_poslist'].lat
	propagationData['o_poslist'].lon
	propagationData['o_poslist'].sun_pos
	propagationData['o_poslist'].moon_pos
	propagationData['o_poslist'].alt
	propagationData['o_poslist'].eclipse
	propagationData['o_poslist'].central_body
	propagationData['o_poslist'].period
	propagationData['o_poslist'].period_steps
	propagationData['o_poslist'].semi_major
	propagationData['o_poslist'].ecc
	propagationData['o_poslist'].inc
	propagationData['o_poslist'].raan
	propagationData['o_poslist'].argp
	assert propagationData['o_poslist'].timespan is not None
	assert propagationData['o_poslist'].name is not None
	assert propagationData['o_poslist'].gen_type is not None
	assert propagationData['o_poslist'].pos is not None
	assert propagationData['o_poslist'].vel is not None

	propagationData['o_poslist_astro'].name
	propagationData['o_poslist_astro'].satcat_id
	propagationData['o_poslist_astro'].gen_type
	propagationData['o_poslist_astro'].timespan
	propagationData['o_poslist_astro'].TLE_epochs
	propagationData['o_poslist_astro'].pos
	propagationData['o_poslist_astro'].pos_ecef
	propagationData['o_poslist_astro'].vel_ecef
	propagationData['o_poslist_astro'].vel
	propagationData['o_poslist_astro'].lat
	propagationData['o_poslist_astro'].lon
	propagationData['o_poslist_astro'].sun_pos
	propagationData['o_poslist_astro'].moon_pos
	propagationData['o_poslist_astro'].alt
	propagationData['o_poslist_astro'].eclipse
	propagationData['o_poslist_astro'].central_body
	propagationData['o_poslist_astro'].period
	propagationData['o_poslist_astro'].period_steps
	propagationData['o_poslist_astro'].semi_major
	propagationData['o_poslist_astro'].ecc
	propagationData['o_poslist_astro'].inc
	propagationData['o_poslist_astro'].raan
	propagationData['o_poslist_astro'].argp
	assert propagationData['o_poslist_astro'].timespan is not None
	assert propagationData['o_poslist_astro'].name is not None
	assert propagationData['o_poslist_astro'].gen_type is not None
	assert propagationData['o_poslist_astro'].pos is not None
	assert propagationData['o_poslist_astro'].vel is not None
	assert propagationData['o_poslist_astro'].sun_pos is not None
	assert propagationData['o_poslist_astro'].moon_pos is not None
	assert propagationData['o_poslist_astro'].eclipse is not None

def test_StaticGenAttributes(propagationData:dict):
	'''
	Test for existance of required attributes in the orbit class
	'''

	propagationData['o_static'].name
	propagationData['o_static'].satcat_id
	propagationData['o_static'].gen_type
	propagationData['o_static'].timespan
	propagationData['o_static'].TLE_epochs
	propagationData['o_static'].pos
	propagationData['o_static'].pos_ecef
	propagationData['o_static'].vel_ecef
	propagationData['o_static'].vel
	propagationData['o_static'].lat
	propagationData['o_static'].lon
	propagationData['o_static'].sun_pos
	propagationData['o_static'].moon_pos
	propagationData['o_static'].alt
	propagationData['o_static'].eclipse
	propagationData['o_static'].central_body
	propagationData['o_static'].period
	propagationData['o_static'].period_steps
	propagationData['o_static'].semi_major
	propagationData['o_static'].ecc
	propagationData['o_static'].inc
	propagationData['o_static'].raan
	propagationData['o_static'].argp
	assert propagationData['o_static'].timespan is not None
	assert propagationData['o_static'].name is not None
	assert propagationData['o_static'].gen_type is not None
	assert propagationData['o_static'].pos is not None
	assert propagationData['o_static'].vel is not None

	propagationData['o_static_astro'].name
	propagationData['o_static_astro'].satcat_id
	propagationData['o_static_astro'].gen_type
	propagationData['o_static_astro'].timespan
	propagationData['o_static_astro'].TLE_epochs
	propagationData['o_static_astro'].pos
	propagationData['o_static_astro'].pos_ecef
	propagationData['o_static_astro'].vel_ecef
	propagationData['o_static_astro'].vel
	propagationData['o_static_astro'].lat
	propagationData['o_static_astro'].lon
	propagationData['o_static_astro'].sun_pos
	propagationData['o_static_astro'].moon_pos
	propagationData['o_static_astro'].alt
	propagationData['o_static_astro'].eclipse
	propagationData['o_static_astro'].central_body
	propagationData['o_static_astro'].period
	propagationData['o_static_astro'].period_steps
	propagationData['o_static_astro'].semi_major
	propagationData['o_static_astro'].ecc
	propagationData['o_static_astro'].inc
	propagationData['o_static_astro'].raan
	propagationData['o_static_astro'].argp
	assert propagationData['o_static_astro'].timespan is not None
	assert propagationData['o_static_astro'].name is not None
	assert propagationData['o_static_astro'].gen_type is not None
	assert propagationData['o_static_astro'].pos is not None
	assert propagationData['o_static_astro'].vel is not None
	assert propagationData['o_static_astro'].sun_pos is not None
	assert propagationData['o_static_astro'].moon_pos is not None

def test_dimensions(propagationData:dict):
# 	'''
# 	Test pos and vel attributes are same dimensions as timespan
# 	'''

	# From TLE
	check.equal(propagationData['o_tle'].pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].vel.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].pos_ecef.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].vel_ecef.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].lat.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].lon.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].alt.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].eclipse.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].TLE_epochs.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle_astro'].sun_pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle_astro'].moon_pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].semi_major.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].ecc.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].inc.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].raan.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_tle'].argp.shape[0], len(propagationData['t']))

	# From fake TLE
	check.equal(propagationData['o_ftle'].pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].vel.shape[0], len(propagationData['t']))
	# TODO: include pos_ecef, vel_ecef, lat, lon once these are included in the generati']on
	check.equal(propagationData['o_ftle'].alt.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].eclipse.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].TLE_epochs.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle_astro'].sun_pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle_astro'].moon_pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].semi_major.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].ecc.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].inc.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].raan.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_ftle'].argp.shape[0], len(propagationData['t']))

	# From orbital param
	check.equal(propagationData['o_analytical'].pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].vel.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].alt.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].eclipse.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical_astro'].sun_pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical_astro'].moon_pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].semi_major.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].ecc.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].inc.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].raan.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_analytical'].argp.shape[0], len(propagationData['t']))

	# From position list
	check.equal(propagationData['o_poslist'].pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_poslist'].vel.shape[0], len(propagationData['t']))

	# From const pos
	check.equal(propagationData['o_static'].pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_static'].vel.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_static_astro'].sun_pos.shape[0], len(propagationData['t']))
	check.equal(propagationData['o_static_astro'].moon_pos.shape[0], len(propagationData['t']))

@pytest.mark.filterwarnings("ignore::erfa.ErfaWarning")
def test_unsafe(propagationData:dict):
	# TLE prior to earliest zarya epoch
	prior_start_t = timespan.TimeSpan(dt.datetime(1997,1,1),'1S','1M')
	# timespan begins more than 14 days after last TLE epoch
	post_start_t = timespan.TimeSpan(dt.datetime.now() + dt.timedelta(days=30),'1S','1M')
	# timespan begins within given TLE epochs, but more than 14 days after last TLE epoch
	# needs to be earlier than 2100 for external library compatibility
	post_end_t = timespan.TimeSpan(dt.datetime(2005,6,1),'365d','34310d')

	with check.raises(exceptions.OutOfRangeError):
		orbit.Orbit.fromTLE(prior_start_t, pathlib.Path(propagationData['ISS_tle_path']))
		orbit.Orbit.fromTLE(post_start_t, pathlib.Path(propagationData['ISS_tle_path']))
		orbit.Orbit.fromTLE(post_end_t, pathlib.Path(propagationData['ISS_tle_path']))

	orbit.Orbit.fromTLE(prior_start_t, pathlib.Path(propagationData['ISS_tle_path']), unsafe=True)
	orbit.Orbit.fromTLE(post_start_t, pathlib.Path(propagationData['ISS_tle_path']), unsafe=True)
	orbit.Orbit.fromTLE(post_end_t, pathlib.Path(propagationData['ISS_tle_path']), unsafe=True)

	with check.raises(exceptions.OutOfRangeError):
		orbit.Orbit.fromPropagatedOrbitalParam(propagationData['t'], a=(6378 - 600), ecc=0, inc=45, raan=0, argp=0, mean_nu=0)
		orbit.Orbit.fromPropagatedOrbitalParam(propagationData['t'], a=(6378 - 600), ecc=0, inc=45, raan=0, argp=0, mean_nu=0)

	orbit.Orbit.fromPropagatedOrbitalParam(propagationData['t'], a=(6378 - 600), ecc=0, inc=45, raan=0, argp=0, mean_nu=0, unsafe=True)
	orbit.Orbit.fromPropagatedOrbitalParam(propagationData['t'], a=(6378 - 600), ecc=0, inc=45, raan=0, argp=0, mean_nu=0, unsafe=True)

def test_analyticalValidity():
		'''
		Tests to check pos and vel values are sensible for an orbit
		'''

		t0 = dt.datetime(2021, 1, 1, 0, 0, 1)
		t = timespan.TimeSpan(t0, '1S', '180M')
		a = 6378+600
		o = orbit.Orbit.fromAnalyticalOrbitalParam(t, a=a, ecc=0, inc=45, raan=0, argp=0, true_nu=0)

		# Circular orbit should have the same semi-major for its duration
		check.is_true(np.all(np.isclose(np.linalg.norm(o.pos, axis=1), a)))
		# Circular orbit should have the same orbital speed for its duration
		speed = orbital_u.calcOrbitalVel(a * 1e3, np.array((a * 1e3, 0, 0)))
		check.is_true(np.all(np.isclose(np.linalg.norm(o.vel, axis=1), speed)))


def test_fromAnalyticalOrbitalParams_epoch():
		'''
		Test that the epoch argument is set to the correct default value
		'''

		t0 = dt.datetime.fromisoformat('2026-03-20T14:46:00+00:00') # this is the vernal equinox so sun should be at raan
		t = timespan.TimeSpan(t0, '1S', '1S')
		o = orbit.Orbit.fromAnalyticalOrbitalParam(t, a=6378+600, ecc=0.4, inc=45, raan=0, argp=0, true_nu=0)

		def unit_vector(v : NDArray) -> NDArray:
			return v/np.linalg.norm(v)

		# an orbit with raan=argp=true_nu=0 with epoch at vernal equinox should be at the sub solar point
		check.is_true(np.all(np.isclose(unit_vector(o.pos[0]), unit_vector(o.sun_pos[0]), atol=0.01)))

def test_fromAnalyiticalOrbitalParams_epoch_timespan_sampling():
		'''
		test that the epoch is kept consistent over different timespans when specified
		'''

		t0 = dt.datetime.fromisoformat('2026-01-12T23:34:18+00:00')

		ts1 = timespan.TimeSpan(t0, '1M', '180M')
		ts2 = timespan.TimeSpan(t0 + dt.timedelta(minutes=1), '2M', '180M')

		o1 = orbit.Orbit.fromAnalyticalOrbitalParam(ts1, a=6378+400, ecc=0.2, inc=10, raan=18, argp=76, true_nu=23, epoch=ts1[0])
		o2 = orbit.Orbit.fromAnalyticalOrbitalParam(ts2, a=6378+400, ecc=0.2, inc=10, raan=18, argp=76, true_nu=23, epoch=ts1[0])


		check.is_true(
			np.all(np.isclose(
					o1.pos[1::2,:],
					o2.pos[:-1,:],
					atol=0.01
				)))



"""Functions to fetch TLEs from Celestrak.

Attributes:
	MAX_RETRIES: number of times to try and reach Celestrak.
	TIMEOUT: timeout for connection and request servicing by Celestrak.
"""

import datetime as dt
import logging
import pathlib

import requests

import spherapy
from spherapy.util import epoch_u

MAX_RETRIES=3
TIMEOUT=10

logger = logging.getLogger(__name__)

def updateTLEs(sat_id_list:list[int]) -> list[int]:
	"""Fetch most recent TLE for satcat IDs from celestrak.

	Fetch most recent TLE for provided list of satcat IDs, and store in file.
	Will try MAX_RETRIES before raising a ValueError

	Args:
		sat_id_list: list of satcat ids to fetch

	Returns:
		list: list of satcat ids successfully fetched

	Raises:
		TimeoutError
	"""
	logger.info("Using CELESTRAK to update TLEs")
	modified_list = []
	for sat_id in sat_id_list:
		url = f'https://celestrak.org/NORAD/elements/gp.php?CATNR={sat_id}'
		retry_num = 0
		fetch_successful = False
		while retry_num < MAX_RETRIES:
			retry_num += 1
			r = requests.get(url, timeout=TIMEOUT)
			if r.status_code == requests.codes.ok:
				fetch_successful = True
				tle_file = getTLEFilePath(sat_id)
				dat_list = r.text.split('\r\n')
				with tle_file.open('w') as fp:
					fp.write(f'0 {dat_list[0].rstrip()}\n')
					fp.write(f'{dat_list[1].rstrip()}\n')
					fp.write(f'{dat_list[2].rstrip()}')
					modified_list.append(sat_id)
				break

		if retry_num == MAX_RETRIES and fetch_successful:
			logger.error('Could not fetch celestrak information for sat_id: %s', sat_id)
			raise TimeoutError(f'Could not fetch celestrak information for sat_id: {sat_id}')

	return modified_list

def getTLEFilePath(sat_id:int) -> pathlib.Path:
	"""Gives path to file where celestrak TLE is stored.

	Celestrak TLEs are stored in {satcadID}.temptle

	Args:
		sat_id: satcat ID

	Returns:
		path to file
	"""
	return spherapy.tle_dir.joinpath(f'{sat_id}.temptle')

def getStoredEpochs(sat_id:int) -> None|tuple[dt.datetime, dt.datetime|None]:
	"""Return the start and end epoch for {sat_id}.temptle .

	Args:
		sat_id: satcat id to check

	Returns:
		(first epoch datetime, last epoch datetime)
		None if no spacetrack tle stored for sat_id
	"""
	tle_path = getTLEFilePath(sat_id)
	return epoch_u.getStoredEpochs(tle_path)



"""Namespace of useful orbital constants."""

import astropy.constants as astroconst

# ######### BODY CONSTANTS ##########
M_EARTH = astroconst.M_earth.value 	# Mass of earth (Kg)
M_SUN = astroconst.M_sun.value 	# Mass of the sun (Kg)
M_MOON = 7.342e22 	# Mass of the sun (Kg)
M_MARS = 6.4171e23 	# Mass of the sun (Kg)
R_SUN = astroconst.R_sun.value / 1000  # Radius of sun (695700 km)
R_EARTH = astroconst.R_earth.value / 1000    # Radius of Earth (6378.1 km)
R_MOON = 1738 	# Radius of the Moon (km)
R_MARS = 3390 	# Radius of Mars (km)
SUN_MIN_ALT = 0 	# Minimum orbital altitude of the sun (km)
EARTH_MIN_ALT = 350 	# Minimum orbital altitude of the earth (km)
MOON_MIN_ALT = 10 	# Minimum orbital altitude of the moon (km)
MARS_MIN_ALT = 200 	# Minimum orbital altitude of mars (km)
G = astroconst.G.value 	# Gravitational constant (SI)
GM_EARTH = astroconst.GM_earth.value 	# Earth standard gravitational parameter (SI)
GM_SUN = astroconst.GM_sun.value 	# Sun standard gravitational parameter (SI)
GM_MOON = 4.9048695e12 	# Moon standard gravitational parameter (SI)
GM_MARS = 4.282837e13 	# Mars standard gravitational parameter (SI)
AU = astroconst.au.value / 1000 	# Earth-Sun avg distance (1.49597871e8 km)
J2 = 1082.6267e-6 	# Earth J2 perturbations (SI)
W_EARTH = 7.29211510e-5 # Rotation rate of Earth rads/sec
# ######### TEMP CONSTANTS ##########
SB_SIGMA = astroconst.sigma_sb.value 	# the Stefan-Boltzmann constant (5.67037442e-8 SI)
T_EARTH = 250     # Temperature of the Earth (K)
T_SUN = 5780    # Temperature of the Sun (K)
T_SPACE = 4       # Temperature of space (K)
SOL_CONST = 1391    # Max. flux of the sun at the earth (W/m^2)
L_SUN     = 3.828e26 # Solar luminosity in watts

# ######### TIME CONSTANTS ##########

DAY_IN_MINS = 24 * 60
HALF_HOUR = 30
SECS_PER_DAY = 24 * 60 * 60



import pytest
import numpy as np
from spherapy.util import epoch_u
import datetime as dt


import logging
logging.disable(logging.CRITICAL)


@pytest.fixture(scope='module')
def date_ranges() -> tuple[np.ndarray, np.ndarray]:
	date_range1 = np.arange(dt.datetime(2020,7,1), dt.datetime(2021,1,1), dt.timedelta(days=0.5)).astype(dt.datetime)
	date_range2 = np.arange(dt.datetime(2020,7,1,0,0,10), dt.datetime(2021,8,1,0,0,10), dt.timedelta(minutes=1)).astype(dt.datetime)
	return date_range1, date_range2

def test_priorToSourceArr(date_ranges:tuple):
	search_arr = np.arange(dt.datetime(2019,7,1), dt.datetime(2019,7,1,5,59), dt.timedelta(minutes=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.zeros(len(search_arr))
	assert np.allclose(idxs, expected_arr)

def test_postSourceArr(date_ranges:tuple):
	search_arr = np.arange(dt.datetime(2022,7,1), dt.datetime(2022,7,1,5,59), dt.timedelta(minutes=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.full(len(search_arr),len(date_ranges[0])-1)
	assert np.allclose(idxs, expected_arr)

def test_withinClosestToFirst(date_ranges:tuple):
	# all closest to idx=0
	search_arr = np.arange(dt.datetime(2020,7,1), dt.datetime(2020,7,1,5,59), dt.timedelta(minutes=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.zeros(len(search_arr))
	assert np.allclose(idxs, expected_arr)

def test_withinSplitFirstSecond1(date_ranges:tuple):
	# all but last closest to idx=0
	search_arr = np.arange(dt.datetime(2020,7,1), dt.datetime(2020,7,1,6,2), dt.timedelta(minutes=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.zeros(len(search_arr))
	expected_arr[-1] = 1
	assert np.allclose(idxs, expected_arr)

def test_withinSplitFirstSecond2(date_ranges:tuple):
	# split between 330 and 331 for idx = 0
	search_arr = np.arange(dt.datetime(2020,7,1,0,30), dt.datetime(2020,7,1,6,32), dt.timedelta(minutes=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.zeros(len(search_arr))
	expected_arr[-31:] = 1
	assert np.allclose(idxs, expected_arr)

def test_withinSplitArbitrary1(date_ranges:tuple):
	# all but last closest to idx=317 (randomly chosen)
	search_arr = np.arange(dt.datetime(2020, 12, 6, 12, 0), dt.datetime(2020, 12, 6, 18, 2), dt.timedelta(minutes=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.full(len(search_arr), 317)
	expected_arr[-1] = 318
	assert np.allclose(idxs, expected_arr)

def test_withinSplitArbitrary2(date_ranges:tuple):
	# split between 270 and 271 for idx = 317
	search_arr = np.arange(dt.datetime(2020, 12, 6, 13, 30), dt.datetime(2020, 12, 6, 19, 32), dt.timedelta(minutes=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.full(len(search_arr), 317)
	expected_arr[-91:] = 318
	assert np.allclose(idxs, expected_arr)

def test_withinMultiSplit1(date_ranges:tuple):
	# closest to 131,132,133,134,135
	search_arr = np.arange(dt.datetime(2020, 9, 4, 9, 5, 25), dt.datetime(2020, 9, 6, 8, 10, 0), dt.timedelta(seconds=477)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.full(len(search_arr), 131)
	expected_arr[68:158] = 132
	expected_arr[158:249] = 133
	expected_arr[249:339] = 134
	expected_arr[339:] = 135
	assert np.allclose(idxs, expected_arr)

def test_withinMultiSplit2(date_ranges:tuple):
	# closest to 131,132,133,134,135
	search_arr = np.arange(dt.datetime(2020, 9, 4, 12, 5, 25), dt.datetime(2020, 9, 6, 11, 10, 0), dt.timedelta(seconds=477)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[0])
	assert len(search_arr) == len(idxs)
	expected_arr = np.full(len(search_arr), 131)
	expected_arr[45:136] = 132
	expected_arr[136:226] = 133
	expected_arr[226:317] = 134
	expected_arr[317:] = 135
	assert np.allclose(idxs, expected_arr)

def test_searchArrStepsLarger(date_ranges:tuple):
	search_arr = np.arange(dt.datetime(2020,7,2), dt.datetime(2020,7,9), dt.timedelta(days=1)).astype(dt.datetime)
	idxs = epoch_u.findClosestDatetimeIndices(search_arr, date_ranges[1])
	assert len(search_arr) == len(idxs)
	expected_arr = np.array((1440, 2880, 4320, 5760, 7200, 8640, 10080))
	assert np.allclose(idxs, expected_arr)



"""Utility functions for orbital calculations."""
import numpy as np

import spherapy.util.constants as consts


def ssoInc(alt:float, e:float=0) -> float:
	# TODO: update to calculate for different central bodies
	"""Generates required inclination for given altitude [km] to maintain Sun Syncrhonous orbit.

	Args:
		alt: altitude of orbit in km
		e: [Optional] eccentricity of orbit
			Default is circular (0)

	Returns:
		Inclination angle in degrees
	"""
	a = consts.R_EARTH + alt * 1e3
	# print(a)
	p = a * (1 - e**2)
	# print(p)
	period = 2 * np.pi * np.sqrt(a**3 / consts.GM_EARTH)
	# print(period)
	req_prec_rate = (2 * np.pi / 365.25) * (1 / 86400.)
	# print(req_prec_rate)
	req_prec_orb = req_prec_rate * period
	# print(req_prec_orb)

	cosi = (req_prec_orb * p**2) / (-3 * np.pi * consts.J2 * consts.R_EARTH**2)
	# print(cosi)

	inc = np.arccos(cosi)

	return np.rad2deg(inc)


def calcPeriod(a:float) -> float:
	"""Returns the period of an elliptical or circular orbit.

	Args:
		a: semi-major axis in m

	Returns:
		Orbital period in s
	"""
	return 2 * np.pi * np.sqrt(a**3 / consts.GM_EARTH)

def calcOrbitalVel(a:float, pos:np.ndarray[tuple[int],np.dtype[np.float64]]) -> float:
	"""Return the instantaneous velocity magnitude for an elliptical orbit at position.

	Args:
		a: semi-major axis in m
		pos: cartesian position, assuming the origin is at the central body.

	Returns:
		instantaneous velocity magnitude.
	"""
	r = np.linalg.norm(pos)

	return np.sqrt(consts.GM_EARTH * (2 / r - 1 / a))

def calcMeanMotion(a:float) -> float:
	"""Returns mean motion [radians/s] for an elliptical or circular orbit with semi-major axis a.

	Args:
		a: semi-major axis in m

	Returns:
		Orbital period in s
	"""
	return np.sqrt(consts.GM_EARTH / a**3)



"""Functions to fetch and store spacetrack credentials on the system."""

import configparser
import getpass

import keyring
import keyring.errors

import spherapy

# used to store/fetch the username in keyring (abuse of API)
USERNAME_KEY = "spherapy_username"

def fetchConfigCredentials(config:configparser.ConfigParser) -> dict[str, str|None]:
	"""Fetch spacetrack credentials from config file.

	Args:
		config: ConfigParser object containing config fields

	Returns:
		dict containing username and password
	"""
	username = config['credentials'].get('SpacetrackUser')
	if username == 'None':
		username = None

	password = config['credentials'].get('SpacetrackPasswd')
	if password == 'None': 		# noqa: S105, password not hardcoded, just parsing empty config file
		password = None

	return {'user':username, 'passwd':password}

def fetchKeyringCredentials() -> dict[str,str|None]:
	"""Fetch spacetrack credentials from system keyring.

	Returns:
		dict containing username and password

	Raises:
		ValueError: If no system keyring exists
	"""
	username = _fetchUser()
	if username is None:
		return {'user':None, 'passwd':None}
	password = _fetchPass(username)
	return {'user':username, 'passwd':password}

def _fetchUser() -> str|None:
	try:
		username = keyring.get_password(spherapy.service_name, USERNAME_KEY)
	except keyring.errors.NoKeyringError:
		raise ValueError('No Keyring exists on this machine: '
							'did you forget to set env variable SPHERAPY_CONFIG_DIR '
							'if using on a headless machine?') \
							from keyring.errors.NoKeyringError
	return username

def _fetchPass(username:str) -> str|None:
	try:
		password = keyring.get_password(spherapy.service_name, username)
	except keyring.errors.NoKeyringError:
		raise ValueError('No Keyring exists on this machine: '
							'did you forget to set env variable SPHERAPY_CONFIG_DIR') \
							from keyring.errors.NoKeyringError
	return password

def _reloadCredentials(config_parser:configparser.ConfigParser|None=None):
	if config_parser is None:
		spherapy.spacetrack_credentials = fetchKeyringCredentials()
	else:
		spherapy.spacetrack_credentials = fetchConfigCredentials(config_parser)

def storeCredentials(user:None|str=None, passwd:None|str=None) -> bool:
	"""Store spacetrack credentials in system keyring.

	Can set either username or password, but can't set a password without a username

	Args:
		user: [optional] username string
		passwd: [optional] password string

	Returns:
		True if successfully stored

	Raises:
		KeyError: raised if trying to store a password without a username
	"""
	if user is not None:
		keyring.set_password(spherapy.service_name, USERNAME_KEY, user)
	else:
		user = _fetchUser()
		if user is None:
			raise KeyError("Can't set a Spacetrack password without a username: "
							"no existing username")
	if passwd is not None:
		keyring.set_password(spherapy.service_name, user, passwd)

	# check stored credentials match input
	stored_user = _fetchUser()
	stored_pass = _fetchPass(stored_user) if stored_user is not None else None
	if (stored_user == user and stored_pass == passwd):
		_reloadCredentials()
		return True

	return False

def clearCredentials() -> None:
	"""Clear the stored spacetrack credentials.

	Raises:
		KeyError: If no credentials are currently stored.
	"""
	user = _fetchUser()
	if user is not None:
		keyring.delete_password(spherapy.service_name, user)
	else:
		raise KeyError("No stored USER - no credentials to delete.")
	keyring.delete_password(spherapy.service_name, USERNAME_KEY)

	_reloadCredentials()

def createCredentials():
	"""Script helper function to create and store credentials.

	Called by command line script. Requires user input.
	"""
	user = input('Please enter your spacetrack username:')
	passwd = getpass.getpass(prompt='Please enter your spacetrack password:')
	print("These credentials are being saved in your system keyring") 	# noqa: T201 print needed
	if not storeCredentials(user=user, passwd=passwd):
		print("Could not save credentials in system keyring") 	# noqa: T201 print needed



"""Generic exceptions that apply throughout satplot."""

class OutOfRangeError(Exception):
	"""The value is out of the acceptable range."""

class DimensionError(Exception):
	"""Dimension of an array is invalid."""



import pytest
import spherapy
from importlib import resources

@pytest.fixture(scope='module')

def pytestFillPackagedData():
	old_packaged_TLEs = spherapy.packaged_TLEs
	print('Filling packaged TLEs index from data dir')  		#noqa: T201

	# find path of spherapy package dir
	package_dir = resources.files('spherapy')
	# packaged TLEs stored in
	TLE_dir = package_dir.joinpath('data').joinpath('TLEs')
	entries = TLE_dir.glob('**/*')
	files = [x for x in entries if x.is_file()]

	packaged_TLEs = {}
	# extract TLE files and add to packaged_TLEs
	for file in files:
		try:
			tle_id = int(file.stem)
		except ValueError:
			print("Can't import packaged TLEs, skipping...") 	#noqa: T201
			return None
		packaged_TLEs[tle_id] = file
	spherapy.packaged_TLEs = packaged_TLEs

	yield
	# Tear down
	print('Restoring packaged TLEs:') 							#noqa: T201
	spherapy.packaged_TLEs = old_packaged_TLEs



"""Class for orbital data.

This module provides:
- OrbitAttrDict: A TypedDict typing the instance attributes of an Orbit
- Orbit: A class of timestamped orbital data
"""
import datetime as dt
import logging
import pathlib

import types
import typing
from typing import Any, TypedDict, cast

from astropy import units as astropy_units
from astropy.time import Time as astropyTime
from hapsira.bodies import Earth, Mars, Moon, Sun
from hapsira.ephem import Ephem
from hapsira.twobody import Orbit as hapsiraOrbit
import numpy as np
from sgp4.api import WGS72, Satrec

# Don't need to import all of skyfield just for EarthSatellite loading.
# TODO: figure out the lightest import to use
from skyfield.api import EarthSatellite, load, wgs84
from skyfield.framelib import itrs

from spherapy.timespan import TimeSpan
from spherapy.util import epoch_u, exceptions
import spherapy.util.constants as consts
import spherapy.util.orbital_u as orbit_u

logger = logging.getLogger(__name__)


class OrbitAttrDict(TypedDict):
	"""A TypedDict providing type annotations for the Orbit class.

	Attributes:
		name:
		satcat_id:
		gen_type:
		timespan:
		TLE_epochs:
		pos:
		pos_ecef:
		vel_ecef:
		vel:
		lat:
		lon:
		sun_pos:
		moon_pos:
		alt:
		eclipse:
		central_body:
		period:
		period_steps:
		semi_major:
		ecc:
		inc:
		raan:
		argp:
	"""
	name: None|str
	satcat_id: None|int
	gen_type: None|str
	timespan: None|TimeSpan
	TLE_epochs: None|np.ndarray[tuple[int], np.dtype[np.datetime64]]
	pos: None|np.ndarray[tuple[int,int], np.dtype[np.float64]]
	pos_ecef: None|np.ndarray[tuple[int,int], np.dtype[np.float64]]
	vel_ecef: None|np.ndarray[tuple[int,int], np.dtype[np.float64]]
	vel: None|np.ndarray[tuple[int,int], np.dtype[np.float64]]
	lat: None|np.ndarray[tuple[int], np.dtype[np.float64]]
	lon: None|np.ndarray[tuple[int], np.dtype[np.float64]]
	sun_pos: None|np.ndarray[tuple[int,int], np.dtype[np.float64]]
	moon_pos: None|np.ndarray[tuple[int,int], np.dtype[np.float64]]
	alt: None|np.ndarray[tuple[int], np.dtype[np.float64]]
	eclipse: None|np.ndarray[tuple[int], np.dtype[np.bool_]]
	central_body: None|str
	period: None|float
	period_steps: None|int
	semi_major: None|np.ndarray[tuple[int], np.dtype[np.float64]]
	ecc: None|np.ndarray[tuple[int], np.dtype[np.float64]]
	inc: None|np.ndarray[tuple[int], np.dtype[np.float64]]
	raan: None|np.ndarray[tuple[int], np.dtype[np.float64]]
	argp: None|np.ndarray[tuple[int], np.dtype[np.float64]]

def _createEmptyOrbitAttrDict() -> OrbitAttrDict:
	return OrbitAttrDict({
	'name': None,
	'satcat_id':None,
	'gen_type':None,
	'timespan':None,
	'TLE_epochs':None,
	'pos':None,
	'pos_ecef':None,
	'vel_ecef':None,
	'vel':None,
	'lat':None,
	'lon':None,
	'sun_pos':None,
	'moon_pos':None,
	'alt':None,
	'eclipse':None,
	'central_body':None,
	'period':None,
	'period_steps':None,
	'semi_major':None,
	'ecc':None,
	'inc':None,
	'raan':None,
	'argp':None,
	})

def _validateOrbitAttrDict(obj:OrbitAttrDict, typ: type) -> bool:
	"""Validates all fields of a TypedDict have been instantiated, and are of correct type.

	Raises:
		KeyError: If a field of the TypedDict is missing
		TypeError: If a field of the TYpedDict is incorrect
	"""
	for attr_name, spec_attr_type in typ.__annotations__.items():
		try:
			attr_value = obj.get(attr_name, None)
		except KeyError as e:
			# Check for missing keys
			raise KeyError('Typed Dict is missing key: %s', attr_name) from e
		if type(spec_attr_type) is types.UnionType:
			# check union of types
			good_type = False
			for sub_type in typing.get_args(spec_attr_type):
				# choose validator
				if sub_type in (None, types.NoneType):
					validator = _NoneValidator
				else:
					validator = _genericValidator
				if validator(attr_value, sub_type):
					good_type = True
					break
			if not good_type:
				raise TypeError(f'{attr_name} is type {type(attr_value)}, '
								f'should be {spec_attr_type}')
		elif not _genericValidator(attr_value, spec_attr_type):
			# check all other types
			raise TypeError(f'{attr_name} is type {type(attr_value)}, should be {spec_attr_type}')
	return True

def _NoneValidator(obj:object, check_type:type) -> bool:  #noqa: ARG001
	return type(obj) is types.NoneType

def _genericValidator(obj:object, check_type:type|types.GenericAlias) -> bool:
	if type(check_type) is types.GenericAlias:
		if type(obj) is check_type.__origin__:
			return True
	elif type(obj) is check_type:
			return True
	return False

class Orbit:
	"""Timestamped orbital data for a satellite, coordinate system depending on the central body.

	Attributes:
		timespan: Timespan over which orbit is to be simulated
		name: Name of the satellite
		satcat_id: NORAD satellite catelogue ID (if generated from a TLE)
		gen_type: How the orbit was generated
		TLE_epochs: Nx1 numpy array of TLE epoch used for propagation at each timestamp
				units: TLE epoch
		pos: Nx3 numpy array of cartesian coordinates of the position of the satellite
			at each timestamp
				units: km
				frame: ECI
		vel: Nx3 numpy array of cartesian velocities of the satellite at each timestamp
				units: m/s
				frame: ECI
		pos_ecef: Nx3 numpy array of cartesian coordinates of the position of the satellite
				at each timestamp
					units: km
					frame: ECEF
		vel_ecef: Nx3 numpy array of cartesian velocities of the satellite at each timestamp
					units: m/s
					frame: ECEF
		lat: Nx1 numpy array of central body latitudes of the satellite at each timestamp
				units: degrees
		lon: Nx1 numpy array of central body longitudes of the satellite at each timestamp
				units: degrees
		sun_pos: Nx3 numpy array of cartesian coordinates of the position of the Sun
				at each timestamp
					units: km
					frame: ECI
		moon_pos: Nx3 numpy array of cartesian coordinates of the position of the Moon
				at each timestamp
					units: km
					frame: ECI
		alt: Nx1 numpy array of altitudes above central body at each timestamp
				units: km
		eclipse: Nx1 numpy array of flag indicating if satellite is eclipsed at each timestamp
					units: km
		central_body: body the satellite is orbiting
		period: orbital period in secs
		period_steps: number of TimeSpan timestamps required to complete an orbit
		semi_major: Nx1 numpy array of orbit semi-major axis calculated at that timestep
						units: km
						will be constant if no orbital maneauvers
		ecc: Nx1 numpy array of orbit eccentricity calculated at that timestep
				units: unitless
				will be constant if no orbital maneauvers
		inc: Nx1 numpy array of orbit inclination calculated at that timestep
				units: degree
				will be constant if no orbital maneauvers
		raan: Nx1 numpy array of orbit RAAN calculated at that timestep
				units: degree
				will be constant if no orbital maneauvers
		argp: Nx1 numpy array of orbit Arg Perigee calculated at that timestep
				units: degree
				will be constant if no orbital maneauvers
	"""

	def __init__(self, data:OrbitAttrDict, calc_astrobodies:bool=False):
		"""The constructor should never be called directly.

		Use one of:
			Orbit.fromTLE()
			Orbit.fromListOfPositions()
			Orbit.fromPropagatedOrbitalParam()
			Orbit.fromAnalyticalOrbitalParam()
		"""
		# Should always be called from a class method, spit error if not.
		if not _validateOrbitAttrDict(data, OrbitAttrDict):
			logger.error("Orbit() should not be called directly, "
						"use one of the fromXX constructors.")
			raise ValueError("Orbit() should not be called directly, "
							"use one of the fromXX constructors")

		self.name = data['name']
		self.satcat_id = data['satcat_id']
		self.gen_type = data['gen_type']

		#time data
		self.timespan = data['timespan']
		self.TLE_epochs = data['TLE_epochs']

		#pos data
		self.pos = data['pos']
		self.pos_ecef = data['pos_ecef']
		self.vel_ecef = data['vel_ecef']
		self.vel = data['vel']
		self.lat = data['lat']
		self.lon = data['lon']
		self.alt = data['alt']

		# These should be None from the constructor
		self.sun_pos = data['sun_pos']
		self.moon_pos = data['moon_pos']
		self.eclipse = data['eclipse']

		#orbit_data
		self.central_body = data['central_body']
		self.period = data['period']
		self.period_steps = data['period_steps']
		self.semi_major = data['semi_major']
		self.ecc = data['ecc']
		self.inc = data['inc']
		self.raan = data['raan']
		self.argp = data['argp']

		# Check required fields are not empty
		if self.timespan is None:
			raise AttributeError('Error in creating Orbit object, no Timespan assigned')
		if self.pos is None:
			raise AttributeError('Error in creating Orbit object, no position data assigned')
		if self.vel is None:
			raise AttributeError('Error in creating Orbit object, no velocity data assigned')
		if self.name is None:
			raise AttributeError('Error in creating Orbit object, no name assigned')
		if self.gen_type is None:
			raise AttributeError('Error in creating Orbit object, no generator type assigned')

		if calc_astrobodies:
			logger.info('Creating ephemeris for Sun using timespan')
			# Timescale for sun position calculation should use TDB, not UTC
			# The resultant difference is likely very small
			ephem_sun = Ephem.from_body(Sun,
										astropyTime(self.timespan.asAstropy(scale='tdb')),
										attractor=Earth)
			sun_pos = ephem_sun.rv()[0]
			self.sun_pos = np.asarray(sun_pos.to(astropy_units.km))

			logger.info('Creating ephemeris for Moon using timespan')
			ephem_moon = Ephem.from_body(Moon,
											astropyTime(self.timespan.asAstropy(scale='tdb')),
											attractor=Earth)
			moon_pos = ephem_moon.rv()[0]
			self.moon_pos = np.asarray(moon_pos.to(astropy_units.km))
			self.eclipse = self._calcEclipse(self.pos, self.sun_pos)

	#pos_list
	@classmethod
	def fromListOfPositions(cls, timespan:TimeSpan,
									positions:np.ndarray[tuple[int, int], np.dtype[np.float64]],
									astrobodies:bool=False) -> 'Orbit':
		"""Create an orbit from a list of positions.

		Creat an obit by explicitly specifying the position of the
		satellite at each point in time. Useful for simplified test cases; but
		may lead to unphysical orbits.

		Args:
			timespan: Timespan over which orbit is to be simulated
			positions: Nx3 numpy array of cartesian coordinates of the position of the satellite
						at each timestamp
							units: km
							frame: ECI
			astrobodies: [Optional] Flag to calculate Sun and Moon positions at timestamps
							Default is False

		Returns:
			satplot.Orbit
		"""
		attr_dct = _createEmptyOrbitAttrDict()
		if len(positions) != len(timespan):
			raise ValueError(f"Number of supplied positions does not match timespan length: "
							f"{len(positions)} =/= {len(timespan)}")

		attr_dct['timespan'] = timespan
		attr_dct['pos'] = positions
		# Assume linear motion between each position at each timestep;
		# Then assume it stops at the last timestep.
		vel = positions[1:] - positions[:-1]
		attr_dct['vel'] = np.concatenate((vel, np.array([[0, 0, 0]])))
		attr_dct['name'] = 'Sat from position list'
		attr_dct['gen_type'] = 'position list'

		return cls(attr_dct, calc_astrobodies=astrobodies)

	#TLE
	@classmethod
	def fromTLE(cls, timespan:TimeSpan,
						tle_path:pathlib.Path,
						astrobodies:bool=True,
						unsafe:bool=False) -> 'Orbit':
		"""Create an orbit from an existing TLE or a list of historical TLEs.

		Args:
			timespan : TimeSpan over which orbit is to be simulated
			tle_path : path to file containing TLEs for a satellite
			astrobodies: [Optional] Flag to calculate Sun and Moon positions at timestamps
							Default is False
			unsafe: [Optional] Flag to ignore TLE usage more than 14 days either side of timestamps
						Optional
						Default is False

		Returns:
		-------
			satplot.Orbit
		"""
		# load all TLEs and create skyfield earth sats
		skyfld_earth_sats = load.tle_file(f'{tle_path}')

		# generate list of epochs for each TLE
		epochs = np.asarray([a.epoch.utc_datetime() for a in skyfld_earth_sats])

		# ensure no duplicate skyfld sats
		_, unq_idxs = np.unique(epochs, return_index=True)
		unq_skyfld_earth_sats = list(np.asarray(skyfld_earth_sats)[unq_idxs])

		skyfld_earth_sats = unq_skyfld_earth_sats
		tle_epoch_dates = [sat.epoch.utc_datetime() for sat in skyfld_earth_sats]

		# check timespan is valid for provided TLEs
		if timespan.start < tle_epoch_dates[0] - dt.timedelta(days=14):
			logger.error("Timespan begins before provided TLEs (+14 days)")
			if not unsafe:
				raise exceptions.OutOfRangeError("Timespan begins before provided TLEs (+14 days)")
		elif timespan.start > tle_epoch_dates[-1] + dt.timedelta(days=14):
			logger.error("Timespan begins after provided TLEs (+14 days)")
			if not unsafe:
				raise exceptions.OutOfRangeError("Timespan begins after provided TLEs (+14 days)")
		elif timespan.end > tle_epoch_dates[-1] + dt.timedelta(days=14):
			logger.error("Timespan ends after provided TLEs (+14 days)")
			if not unsafe:
				raise exceptions.OutOfRangeError("Timespan ends after provided TLEs (+14 days)")

		attr_dct = _createEmptyOrbitAttrDict()

		_timespan_arr = timespan.asDatetime()
		_timespan_arr = cast("np.ndarray[tuple[int], np.dtype[np.datetime64]]", _timespan_arr)
		closest_tle_epochs = epoch_u.findClosestDatetimeIndices(_timespan_arr,
																	np.asarray(tle_epoch_dates))

		d = np.hstack((1, np.diff(closest_tle_epochs)))

		# timestep idx where TLE epoch changes, append len(timespan)
		# to not require separate loop iteration to handle last element
		timespan_epoch_trans_idxs = np.hstack((np.where(d!=0)[0],len(timespan)))
		# tle_dates idx where TLE epoch changes in timespan
		timespan_epoch_trans_tle_idxs = closest_tle_epochs[np.where(d!=0)[0]]

		tle_epoch_idxs = []
		sub_timespans = []
		for ii, trans_start_idx in enumerate(timespan_epoch_trans_idxs[:-1]):
			tle_epoch_idxs.append(timespan_epoch_trans_tle_idxs[ii])
			sub_timespans.append(timespan[trans_start_idx,timespan_epoch_trans_idxs[ii+1]])

		skyfld_ts = load.timescale(builtin=True)

		# Calculate data for first run
		sub_timespan = sub_timespans[0]
		# help out static type analysis (won't be None), no runtime effect
		sub_timespan = cast("np.ndarray[tuple[int], np.dtype[np.datetime64]]",sub_timespan)
		sub_skyfld_earthsat = skyfld_earth_sats[tle_epoch_idxs[0]]
		tle_epoch = tle_epoch_dates[tle_epoch_idxs[0]]
		sat_rec = sub_skyfld_earthsat.at(skyfld_ts.utc(sub_timespan))
		pos = sat_rec.position.km.T
		vel = sat_rec.velocity.km_per_s.T * 1000
		ecef = sat_rec.frame_xyz_and_velocity(itrs)
		pos_ecef = ecef[0].km.T
		vel_ecef = ecef[1].km_per_s.T * 1000

		skyfld_lat, skyfld_lon = wgs84.latlon_of(sat_rec)
		lat = skyfld_lat.degrees
		lon = skyfld_lon.degrees
		ecc = np.tile(sub_skyfld_earthsat.model.ecco,len(sub_timespan))
		inc = np.tile(sub_skyfld_earthsat.model.inclo,len(sub_timespan))
		semi_major = np.tile(sub_skyfld_earthsat.model.a * consts.R_EARTH,len(sub_timespan))
		raan = np.tile(sub_skyfld_earthsat.model.nodeo,len(sub_timespan))
		argp = np.tile(sub_skyfld_earthsat.model.argpo,len(sub_timespan))
		TLE_epochs = np.tile(tle_epoch,len(sub_timespan)) 		#noqa: N806


		# fill in data for any other sub timespans
		for ii in range(1,len(sub_timespans)):
			sub_timespan = sub_timespans[ii]
			# help out static type analysis (won't be None), no runtime effect
			sub_timespan = cast("np.ndarray[tuple[int], np.dtype[np.datetime64]]",sub_timespan)
			sub_skyfld_earthsat = skyfld_earth_sats[tle_epoch_idxs[ii]]
			tle_epoch = tle_epoch_dates[tle_epoch_idxs[ii]]
			sat_rec = sub_skyfld_earthsat.at(skyfld_ts.utc(sub_timespan))
			pos = np.vstack((pos, sat_rec.position.km.T))
			vel = np.vstack((vel, sat_rec.velocity.km_per_s.T * 1000))
			ecef = sat_rec.frame_xyz_and_velocity(itrs)
			pos_ecef = np.vstack((pos_ecef, ecef[0].km.T))
			vel_ecef = np.vstack((vel_ecef, ecef[1].km_per_s.T * 1000))

			skyfld_lat, skyfld_lon = wgs84.latlon_of(sat_rec)
			lat = np.concatenate((lat, skyfld_lat.degrees))
			lon = np.concatenate((lon, skyfld_lon.degrees))
			ecc = np.concatenate((ecc, np.tile(sub_skyfld_earthsat.model.ecco, len(sub_timespan))))
			inc = np.concatenate((inc,
									np.tile(sub_skyfld_earthsat.model.inclo, len(sub_timespan))))
			semi_major = np.concatenate((semi_major,
											np.tile(sub_skyfld_earthsat.model.a * consts.R_EARTH, len(sub_timespan)))) #noqa:E501
			raan = np.concatenate((raan,
									np.tile(sub_skyfld_earthsat.model.nodeo, len(sub_timespan))))
			argp = np.concatenate((argp,
									np.tile(sub_skyfld_earthsat.model.argpo, len(sub_timespan))))
			TLE_epochs = np.concatenate((TLE_epochs, np.tile(tle_epoch, len(sub_timespan)))) 		#noqa: N806

		attr_dct['timespan'] = timespan
		attr_dct['satcat_id'] = int(skyfld_earth_sats[0].target_name.split('#')[-1].split(' ')[0])
		attr_dct['gen_type'] = 'propagated from TLE'
		attr_dct['central_body'] = 'Earth'
		attr_dct['pos'] = pos
		attr_dct['alt'] = np.linalg.norm(pos, axis=1) - consts.R_EARTH
		attr_dct['pos_ecef'] = pos_ecef
		attr_dct['vel_ecef'] = vel_ecef
		attr_dct['vel'] = vel
		attr_dct['lat'] = lat
		attr_dct['lon'] = lon
		attr_dct['ecc'] = ecc
		attr_dct['inc'] = inc
		attr_dct['semi_major'] = semi_major
		attr_dct['raan'] = raan
		attr_dct['argp'] = argp
		attr_dct['TLE_epochs'] = TLE_epochs

		period = float(2 * np.pi / sub_skyfld_earthsat.model.no_kozai * 60)
		attr_dct['period'] = period

		if timespan.time_step is not None:
			attr_dct['period_steps'] = int(period / timespan.time_step.total_seconds())
		else:
			attr_dct['period_steps'] = None
		attr_dct['name'] = sub_skyfld_earthsat.name

		return cls(attr_dct, calc_astrobodies=astrobodies)

	#fake TLE
	@classmethod
	def fromPropagatedOrbitalParam(cls, timespan:TimeSpan, 			#noqa: PLR0913
										a:float=6978,
										ecc:float=0,
										inc:float=0,
										raan:float=0,
										argp:float=0,
										mean_nu:float=0,
										name:str='Fake TLE',
										astrobodies:bool=True,
										unsafe:bool=False) -> 'Orbit':
		"""Create an orbit from orbital parameters, propagated using sgp4.

		Orbits created using this class method will respect gravity corrections such as J4,
		allowing for semi-analytical sun-synchronous orbits.

		Args:
			timespan: Timespan over which orbit is to be simulated
			a: [Optional] semi-major axis of the orbit in km
					Default is 6978 ~ 600km	above the earth.
			ecc: [Optional] eccentricty, dimensionless number 0 < ecc < 1
					Default is 0, which is a circular orbit
			inc: [Optional] inclination of orbit in degrees
					Default is 0, which represents an orbit around the Earth's equator
			raan: [Optional] right-ascension of the ascending node
					Default is 0
			argp: [Optional] argument of the perigee in degrees
					Default is 0, which	represents an orbit with its semimajor axis in
					the plane of the Earth's equator
			mean_nu: [Optional] mean anomaly in degrees
					Default is 0, which represents an orbit that is beginning at periapsis
			name: [Optional] string giving the name of the orbit
					Default is 'Fake TLE'
			astrobodies: [Optional] Flag to calculate Sun and Moon positions at timestamps
					Default is False
			unsafe: [Optional] Flag to ignore semi-major axis inside Earth's radius
					Default is False

		Returns:
			satplot.Orbit
		"""
		if a < consts.R_EARTH + consts.EARTH_MIN_ALT:
			logger.error("Semimajor axis, %s, is too close to Earth", a)
			if not unsafe:
				raise exceptions.OutOfRangeError(f"Semimajor axis, {a}, is too close to Earth")

		if ecc > 1 or ecc < 0:
			logger.error("Eccentricity, %s, is non circular or eliptical", ecc)
			raise exceptions.OutOfRangeError(f"Eccentricity, {ecc}, is non circular or eliptical")

		if inc > 180 or inc < -180: 			#noqa: PLR2004
			logger.error("Inclination, %s, is out of range, should be -180 < inc < 180", inc)
			raise exceptions.OutOfRangeError(f"Inclination, {inc}, is out of range, "
										f"should be -180 < inc < 180")

		if raan > 360 or raan < 0: 				#noqa: PLR2004
			logger.error("RAAN, %s, is out of range, should be 0 < inc < 360", raan)
			raise exceptions.OutOfRangeError(f"RAAN, {raan}, is out of range, "
												f"should be 0 < inc < 360")

		if argp > 360 or argp < 0: 				#noqa: PLR2004
			logger.error("Argument of periapsis, %s, is out of range, "
						"should be 0 < argp < 360", inc)
			raise exceptions.OutOfRangeError(f"Argument of periapsis, {inc}, is out of range, "
										f"should be 0 < argp < 360")

		if mean_nu > 360 or mean_nu < 0: 		#noqa: PLR2004
			logger.error("Mean anomaly, %s, is out of range, should be 0 < mean_nu < 360", mean_nu)
			raise exceptions.OutOfRangeError(f"Mean anomaly, {mean_nu}, is out of range, "
										f"should be 0 < mean_nu < 360")

		attr_dct = _createEmptyOrbitAttrDict()

		skyfld_ts = load.timescale(builtin=True)

		t0_epoch = epoch_u.datetime2sgp4epoch(timespan.start)
		satrec = Satrec()
		mean_motion = orbit_u.calcMeanMotion(a * 1e3)

		satrec.sgp4init(WGS72,		  # gravity model
						'i',  # mode
						1,  # satnum
						t0_epoch,  # mode
						0,  # bstar [/earth radii]
						0,  # ndot [revs/day]
						0,  # nddot [revs/day^3]
						ecc,  # ecc
						argp,  # arg perigee [radians]
						inc,  # inclination [radians]
						mean_nu,  # mean anomaly [radians]
						mean_motion * 60,  # mean motion [rad/min]
						raan)  # raan [radians]

		skyfld_earthsat = EarthSatellite.from_satrec(satrec, skyfld_ts)

		pos = np.empty((len(timespan), 3))
		vel = np.empty((len(timespan), 3))

		_timespan_arr = timespan.asDatetime()
		_timespan_arr = cast("np.ndarray[tuple[int], np.dtype[np.datetime64]]", _timespan_arr)
		for ii, timestep in enumerate(_timespan_arr):
			pos[ii, :] = skyfld_earthsat.at(skyfld_ts.utc(timestep)).position.km
			vel[ii, :] = skyfld_earthsat.at(skyfld_ts.utc(timestep)).velocity.km_per_s * 1000

		# TODO: calculate ecef and lat,lon values.
		# TODO: satrec doesn't have a frame_xyz_and_velocity in this case

		period = float(2 * np.pi / skyfld_earthsat.model.no_kozai * 60)
		if timespan.time_step is not None:
			period_steps = int(period / timespan.time_step.total_seconds())
		else:
			period_steps = None

		attr_dct['name'] = name
		attr_dct['timespan'] = timespan
		attr_dct['gen_type'] = 'propagated from orbital param'
		attr_dct['central_body'] = 'Earth'
		attr_dct['pos'] = pos
		attr_dct['alt'] = np.linalg.norm(pos, axis=1) - consts.R_EARTH
		attr_dct['vel'] = vel
		attr_dct['ecc'] = np.full(len(timespan), ecc)
		attr_dct['inc'] = np.full(len(timespan), inc)
		attr_dct['semi_major'] = np.full(len(timespan), a)
		attr_dct['raan'] = np.full(len(timespan), raan)
		attr_dct['argp'] = np.full(len(timespan), argp)
		attr_dct['TLE_epochs'] = np.full(len(timespan),timespan.start)
		attr_dct['period'] = period
		attr_dct['period_steps'] = period_steps

		return cls(attr_dct, calc_astrobodies=astrobodies)

	#analytical
	@classmethod
	def fromAnalyticalOrbitalParam(cls, timespan:TimeSpan, 					#noqa: C901, PLR0912, PLR0913
											body:str='Earth',
											a:float=6978,
											ecc:float=0,
											inc:float=0,
											raan:float=0,
											argp:float=0,
											true_nu:float=0,
											epoch:dt.datetime|None=None,
											name:str='Analytical',
											astrobodies:bool=True,
											unsafe:bool=False) -> 'Orbit':
		"""Create an analytical orbit defined by orbital parameters.

		Orbits created using this class method will NOT respect gravity corrections such as J4,
		and as such sun-synchronous orbit are not possible.

		Args:
			timespan: Timespan over which orbit is to be simulated
			body: [Optional] string indicating around what body the satellite orbits,
					Default is 'Earth'
					Options are ['Earth','Sun','Mars','Moon']
			a: [Optional] semi-major axis of the orbit in km
					Default is 6978 ~ 600km	above the earth.
			ecc: [Optional] eccentricty, dimensionless number 0 < ecc < 1
					Default is 0, which is a circular orbit
			inc: [Optional] inclination of orbit in degrees
					Default is 0, which represents an orbit around the Earth's equator
			raan: [Optional] right-ascension of the ascending node
					Default is 0
			argp: [Optional] argument of the perigee in degrees
					Default is 0, which	represents an orbit with its semimajor axis in
					the plane of the Earth's equator
			true_nu: [Optional] true anomaly at epoch in degrees
					Default is 0, which represents an orbit that is beginning
					at periapsis on the epoch
			epoch: [Optional] epoch for orbit
					Default is the first timestep of the timespan
			name: [Optional] string giving the name of the orbit
					Default is 'Analytical'
			astrobodies: [Optional] Flag to calculate Sun and Moon positions at timestamps
					Default is False
			unsafe: [Optional] Flag to ignore semi-major axis inside Earth's radius
					Default is False

		Returns:
			satplot.Orbit
		"""
		if body.upper() == 'EARTH':
			central_body = Earth
			min_a = consts.R_EARTH + consts.EARTH_MIN_ALT
		elif body.upper() == 'SUN':
			central_body = Sun
			min_a = consts.R_SUN + consts.SUN_MIN_ALT
		elif body.upper() == 'MARS':
			central_body = Mars
			min_a = consts.R_MARS + consts.MARS_MIN_ALT
		elif body.upper() == 'MOON':
			central_body = Moon
			min_a = consts.R_MOON + consts.MOON_MIN_ALT
		else:
			logger.error("Invalid central body %s. Valid options are Earth, Sun, Mars, Moon", body)
			raise ValueError(f"Invalid central body {body}.")

		if a < min_a:
			logger.error("Semimajor axis, %s, is too close to the central body, {body.upper()}", a)
			if not unsafe:
				raise exceptions.OutOfRangeError(f"Semimajor axis, {a}, is too close "
											f"to the central body, {body.upper()}")

		if ecc > 1 or ecc < 0:
			logger.error("Eccentricity, %s, is non circular or eliptical", ecc)
			raise exceptions.OutOfRangeError(f"Eccentricity, {ecc}, is non circular or eliptical")

		if inc > 180 or inc < -180: 			#noqa: PLR2004
			logger.error("Inclination, %s, is out of range, should be -180 < inc < 180", inc)
			raise exceptions.OutOfRangeError(f"Inclination, {inc}, is out of range, "
										f"should be -180 < inc < 180")

		if raan > 360 or raan < 0: 				#noqa: PLR2004
			logger.error("RAAN, %s, is out of range, should be 0 < inc < 360", inc)
			raise exceptions.OutOfRangeError(f"RAAN, {inc}, is out of range, "
												f"should be 0 < inc < 360")

		if argp > 360 or argp < 0: 				#noqa: PLR2004
			logger.error("Argument of periapsis, %s, is out of range, "
							"should be 0 < argp < 360", raan)
			raise exceptions.OutOfRangeError(f"Argument of periapsis, {raan}, is out of "
										f"range, should be 0 < argp < 360")

		if true_nu > 360 or true_nu < 0: 		#noqa: PLR2004
			logger.error("True anomaly, %s, is out of range, should be 0 < true_nu < 360", true_nu)
			raise exceptions.OutOfRangeError(f"True anomaly, {true_nu}, is out of range, "
										f"should be 0 < true_nu < 360")

		if epoch is None:
			epoch = timespan[0]
		if isinstance(epoch, dt.datetime):
			astropy_epoch = astropyTime(epoch, scale='utc')
		else:
			logger.error("epoch, %s must be of type dt.datetime|astropyTime but"
											" is of type %s", epoch, type(epoch))
			raise TypeError(f"epoch, {epoch} must be of type "
						f"dt.datetime|astropyTime but is of type {type(epoch)}")

		attr_dct = _createEmptyOrbitAttrDict()



		logger.info("Creating analytical orbit")
		orb = hapsiraOrbit.from_classical(central_body,
											a * astropy_units.one * astropy_units.km,
											ecc * astropy_units.one,
											inc * astropy_units.one * astropy_units.deg,
											raan * astropy_units.one * astropy_units.deg,
											argp * astropy_units.one * astropy_units.deg,
											true_nu * astropy_units.one * astropy_units.deg,
											astropy_epoch
											)


		logger.info("Creating ephemeris for orbit, using timespan")
		ephem = Ephem.from_orbit(orb, timespan.asAstropy())

		pos = np.asarray(ephem.rv()[0], dtype=np.float64)
		vel = np.asarray(ephem.rv()[1], dtype=np.float64) * 1000

		period = float(orb.period.unit.in_units('s') * orb.period.value)
		if timespan.time_step is not None:
			period_steps = int(period / timespan.time_step.total_seconds())
		else:
			period_steps = None

		attr_dct['name'] = name
		attr_dct['timespan'] = timespan
		attr_dct['gen_type'] = 'analytical orbit'
		attr_dct['central_body'] = body
		attr_dct['pos'] = pos
		# TODO: altitude should be sourced from central body
		attr_dct['alt'] = np.linalg.norm(pos, axis=1) - consts.R_EARTH
		attr_dct['vel'] = vel
		attr_dct['ecc'] = ecc * np.ones(len(timespan))
		attr_dct['inc'] = inc * np.ones(len(timespan))
		attr_dct['semi_major'] = a * np.ones(len(timespan))
		attr_dct['raan'] = raan * np.ones(len(timespan))
		attr_dct['argp'] = argp * np.ones(len(timespan))
		attr_dct['period'] = period
		attr_dct['period_steps'] = period_steps

		return cls(attr_dct, calc_astrobodies=astrobodies)

	@classmethod
	def fromDummyConstantPosition(cls, timespan:TimeSpan,
							pos:tuple[float,float,float]|np.ndarray[tuple[int],np.dtype[np.float64]],
							sun_pos:tuple[float,float,float]|np.ndarray[tuple[int],np.dtype[np.float64]]|None=None,
							moon_pos:tuple[float,float,float]|np.ndarray[tuple[int],np.dtype[np.float64]]|None=None) -> 'Orbit': 	#noqa: E501
		"""Creates an static orbit for testing.

		Satellite position is defined by pos, while sun and moon positions are optional,
		but can also be specified.

		Args:
			timespan: Timespan over which orbit is to be simulated
			pos: Nx3 numpy array of cartesian coordinates of the position of the satellite
				at each timestamp
					units: km
					frame: ECI
			sun_pos: [Optional] Nx3 numpy array of cartesian coordinates of the position of the Sun
					at each timestamp
						units: km
						frame: ECI
			moon_pos: [Optional] Nx3 numpy array of cartesian coordinates of the position of the
						Moon at each timestamp
							units: km
							frame: ECI

		Returns:
			satplot.Orbit
		"""
		attr_dct = _createEmptyOrbitAttrDict()
		attr_dct['timespan'] = timespan
		attr_dct['name'] = 'dummy_const_pos'
		attr_dct['pos'] = np.full((len(timespan), 3), pos)
		attr_dct['vel'] = np.full((len(timespan), 3), 0)
		attr_dct['gen_type'] = 'constant position for testing'
		obj = cls(attr_dct, calc_astrobodies=False)
		if sun_pos is not None:
			obj.sun_pos = np.full((len(timespan), 3), sun_pos)
		if moon_pos is not None:
			obj.moon_pos = np.full((len(timespan), 3), moon_pos)
		return obj

	def getPosition(self, search_time:dt.datetime|astropyTime) \
							-> np.ndarray[tuple[int],np.dtype[np.float64]]:
		"""Return the position at the specified or closest time.

		Args:
			search_time: the timestamp to search for

		Returns:
			position

		Raises:
			ValueError: orbit has no pos data
		"""
		if self.pos is None:
			raise ValueError("Orbit has no pos data")
		return self._getAttributeClosestTime(self.pos, search_time)

	def getVelocity(self, search_time:dt.datetime|astropyTime) \
							-> np.ndarray[tuple[int], np.dtype[np.float64]]:
		"""Return the velocity at the specified or closest time.

		Args:
			search_time: the timestamp to search for

		Returns:
			velocity

		Raises:
			ValueError: orbit has no vel data
		"""
		if self.vel is None:
			raise ValueError("Orbit has no vel data")
		return self._getAttributeClosestTime(self.vel, search_time)

	def _getAttributeClosestTime(self, attr:np.ndarray[Any, np.dtype[np.float64]],
										search_time:dt.datetime|astropyTime)\
										-> np.ndarray[tuple[int], np.dtype[np.float64]]:

		if self.timespan is None:
			raise ValueError("Orbit has no timespan")

		if isinstance(search_time, dt.datetime):
			_, closest_idx = self.timespan.getClosest(search_time)
		elif isinstance(search_time, astropyTime):
			search_dt = search_time.to_datetime()
			_, closest_idx = self.timespan.getClosest(search_dt)
		else:
			logger.error("%s cannot be used to index %s", search_time, attr)
			raise TypeError(f"{search_time} cannot be used to index {attr}")

		return attr[closest_idx, :]

	def _calcEclipse(self, pos:np.ndarray[tuple[int,int],np.dtype[np.float64]],
							sun:np.ndarray[tuple[int,int],np.dtype[np.float64]])\
							-> np.ndarray[tuple[int],np.dtype[np.bool_]]:
		earth_ang_size = np.arctan(consts.R_EARTH/consts.AU)
		neg_sun_norm = np.linalg.norm(-sun,axis=1)
		neg_sun_unit = -sun/neg_sun_norm[:,None]
		pos_neg_sun = pos-sun
		pos_neg_sun_norm = np.linalg.norm(pos_neg_sun,axis=1)
		pos_neg_sun_unit = pos_neg_sun/pos_neg_sun_norm[:,None]
		earth_sat_ang_sep = np.arccos(np.sum(neg_sun_unit*pos_neg_sun_unit, axis=1))
		sunlit_angsep_truth = earth_sat_ang_sep > earth_ang_size
		sunlit_dist_truth = pos_neg_sun_norm < neg_sun_norm
		sunlit = np.logical_or(sunlit_angsep_truth, sunlit_dist_truth)

		return np.logical_not(sunlit)



# def _findClosestEpochIndices(target:list[float], values:float) -> list[int]:
# 	right_idx = np.searchsorted(target, values)
# 	left_idx = right_idx - 1
# 	# replace any idx greater than len of target with last idx in target
# 	right_idx[np.where(right_idx==len(target))] = len(target) - 1
# 	stacked_idx = np.hstack((left_idx.reshape(-1,1),right_idx.reshape(-1,1)))
# 	target_vals_at_idxs = target[stacked_idx]
# 	closest_idx_columns = np.argmin(np.abs(target_vals_at_idxs - values.reshape(-1,1)),axis=1)
# 	return stacked_idx[range(len(stacked_idx)),closest_idx_columns]



"""Class for series of timestamps.

This module provides:
- TimeSpan: a series of timestamps to be used by an orbit object
"""
import datetime as dt
import logging

from typing import cast, overload
from typing_extensions import Self

from astropy.time import Time as astropyTime
from dateutil.relativedelta import relativedelta
import numpy as np
from skyfield.api import load
import skyfield.timelib

logger = logging.getLogger(__name__)


class TimeSpan:
	"""A series of timestamps.

	Attributes:
		start: The first timestamp
		end: The last timestamp
		time_step: The difference between timestamps in seconds
						Can be None if irregular steps
		time_period: The difference between end and start in seconds

	"""
	def __init__(self, t0:dt.datetime, timestep:str='1S', timeperiod:str='10S'):
		"""Creates a series of timestamps in UTC.

		Difference between each timestamp = timestep
		Total duration = greatest integer number of timesteps less than timeperiod
			If timeperiod is an integer multiple of timestep,
			then TimeSpan[-1] - TimeSpan[0] = timeperiod
			If timeperiod is NOT an integer multiple of timestep,
			then TimeSpan[-1] = int(timeperiod/timestep) (Note: int cast, not round)

		Does not account for Leap seconds, similar to all Posix compliant UTC based time
			representations. see: https://numpy.org/doc/stable/reference/arrays.datetime.html#datetime64-shortcomings
			for equivalent shortcomings.
		Always contains at least two timestamps

		Parameters
		t0: datetime defining the start of the TimeSpan.
				If timezone naive, assumed to be in UTC
				If timezone aware, will be converted to UTC
		timestep: String describing the time step of the time span.
					The string is constructed as an integer or float, followed by a time unit:
					(d)ays, (H)ours, (M)inutes, (S)econds, (mS) milliseconds, (uS) microseconds
					(the default is '1S')
		timeperiod:	String describing the time period of the time span.
					The string is constructed as an integer or float, followed by a time unit:
					(d)ays, (H)ours, (M)inutes, (S)econds, (mS) milliseconds, (uS) microseconds
					(the default is '1d')

		Raises:
		------
		ValueError
		"""
		# convert and/or add timezone info
		if t0.tzinfo is not None:
			t0 = t0.astimezone(dt.timezone.utc)
		else:
			t0 = t0.replace(tzinfo=dt.timezone.utc)

		self.start:dt.datetime = t0
		self.time_step:None|dt.timedelta = self._parseTimestep(timestep)
		self.end:dt.datetime = self._parseTimeperiod(t0, timeperiod)
		self.time_period = self.end - self.start

		logger.info("Creating TimeSpan between %s and %s with %s seconds timestep",
						self.start, self.end, self.time_step.total_seconds())


		if self.start >= self.end:
			logger.error("Timeperiod: %s results in an end date earlier than the start date",
							timeperiod)
			raise ValueError(f"Invalid timeperiod value: {timeperiod}")

		if self.time_period < self.time_step:
			logger.error("Timeperiod: %s is smaller than the timestep %s.", timeperiod, timestep)
			raise ValueError(f"Invalid timeperiod value, too small: {timeperiod}")

		# Calculate step numbers and correct for non integer steps.
		n_down = int(np.floor(self.time_period / self.time_step))
		if n_down != self.time_period / self.time_step:
			logger.info("Rounding to previous integer number of timesteps")
			self.end = n_down * self.time_step + self.start
			self.time_period = self.end - self.start

			logger.info("Adjusting TimeSpan to %s -> %s with %s seconds timestep",
							self.start, self.end, self.time_step.total_seconds())


		self._timearr = np.arange(self.start.replace(tzinfo=None),
									self.end.replace(tzinfo=None)+self.time_step,
									self.time_step).astype(dt.datetime)
		self._timearr = np.vectorize(lambda x: x.replace(tzinfo=dt.timezone.utc))(self._timearr)
		self._skyfield_timespan = load.timescale()


	def __hash__(self) -> int:
		"""Returns a hash of the timespan."""
		return hash((self.start, self.end, self.time_step))

	# Make it callable and return the data for that entry
	def __call__(self) -> np.ndarray[tuple[int], np.dtype[np.datetime64]]:
		"""Returns the internal _timearr when the TimeSpan is called."""
		return self._timearr

	@overload
	def __getitem__(self, idx: None) -> np.ndarray[tuple[int], np.dtype[np.datetime64]]: ...

	@overload
	def __getitem__(self, idx: int | np.integer) -> dt.datetime: ...

	@overload
	def __getitem__(self, idx: slice | list[int] | tuple[int, ...] | np.ndarray) \
						-> np.ndarray[tuple[int], np.dtype[np.datetime64]]: ...

	def __getitem__(self, idx:None|int|np.integer|tuple|list|np.ndarray|slice=None) \
						-> None|dt.datetime|np.ndarray[tuple[int], np.dtype[np.datetime64]]: 	# noqa: PLR0911
		"""Returns an index or slice of the TimeSpan as an array of datetime objects."""
		if idx is None:
			return self._timearr
		if isinstance(idx, int|np.integer):
			return self._timearr[idx]
		if isinstance(idx, tuple):
			if len(idx) == 2: 								#noqa: PLR2004
				return self._timearr[idx[0]:idx[1]]
			return self._timearr[idx[0]:idx[1]:idx[2]]
		if isinstance(idx, list):
			return self._timearr[idx]
		if isinstance(idx, np.ndarray):
			return self._timearr[idx]
		if isinstance(idx,slice):
			return self._timearr[idx]
		raise TypeError('index is unknown type')

	def __eq__(self, other:object) -> bool:
		"""Checks if other TimeSpan is equal to self."""
		if not isinstance(other, TimeSpan):
			return NotImplemented

		if len(self) != len(other):
			return False

		return bool(np.all(self.asDatetime() == other.asDatetime()))

	def __add__(self, other:Self) -> 'TimeSpan':
		"""Apeends other TimeSpan to this one, does not order timesteps."""
		self_copy = TimeSpan(self.start)

		self_copy.start = self.start
		self_copy.time_step = None

		self_copy.end = other.end
		self_copy.time_period = other.end - self_copy.start

		self_copy._timearr = np.hstack((self._timearr, other._timearr))

		return self_copy

	def asAstropy(self, idx:None|int=None, scale:str='utc') -> astropyTime:
		"""Return ndarray of TimeSpan as astropy.time objects.

		Args:
			idx: timestamp index to return inside an astropyTime object
					if no index supplied, returns all timestamps
			scale: astropy time scale, can be one of
					('tai', 'tcb', 'tcg', 'tdb', 'tt', 'ut1', 'utc')

		Returns:
		-------
		ndarray
		"""
		if idx is None:
			return astropyTime(self._timearr, scale=scale)
		return astropyTime(self._timearr[idx], scale=scale)

	def asDatetime(self, idx:None|int=None) \
		-> dt.datetime|np.ndarray[tuple[int], np.dtype[np.datetime64]]:
		"""Return ndarray of TimeSpan as datetime objects.

		Args:
			idx: timestamp index to return as a datetime
					If no index supplied, returns whole array

		Returns:
		-------
		ndarray
		"""
		if idx is None:
			return self._timearr
		return self._timearr[idx]

	def asSkyfield(self, idx:int) -> skyfield.timelib.Time:
		"""Return TimeSpan element as Skyfield Time object.

		Args:
			idx: timestamp index to return as skyfield time object

		Returns:
		-------
		Skyfield Time
		"""
		datetime = self._timearr[idx]
		datetime = datetime.replace(tzinfo=dt.timezone.utc)
		return self._skyfield_timespan.from_datetime(datetime)

	def asText(self, idx:int) -> str:
		"""Returns a text representation of a particular timestamp.

		Timestamp will be formatted as YYYY-mm-dd HH:MM:SS

		Args:
			idx: timestamp index to format

		Returns:
			str:
		"""
		return self._timearr[idx].strftime("%Y-%m-%d %H:%M:%S")

	def secondsSinceStart(self) -> np.ndarray[tuple[int], np.dtype[np.float64]]:
		"""Return ndarray with the seconds of all timesteps since the beginning.

		Returns:
			array of seconds since start for each timestamp
		"""
		diff = self._timearr - self.start
		days = np.vectorize(lambda x: x.days)(diff)
		secs = np.vectorize(lambda x: x.seconds)(diff)
		usecs = np.vectorize(lambda x: x.microseconds)(diff)
		return days * 86400 + secs + usecs * 1e-6

	def getClosest(self, t_search:dt.datetime) -> tuple[dt.datetime, int]:
		"""Find the closest time in a TimeSpan.

		Parameters
		----------
		t_search : datetime to search for in TimeSpan
				If timezone naive, assumed to be in UTC
				If timezone aware, will be converted to UTCtime to find

		Returns:
		-------
		datetime, int
			Closest datetime in TimeSpan, index of closest date in TimeSpan
		"""
		# convert and/or add timezone info
		if t_search.tzinfo is not None:
			t_search = t_search.astimezone(dt.timezone.utc)
		else:
			t_search = t_search.replace(tzinfo=dt.timezone.utc)
		diff = self._timearr - t_search
		out = np.abs(np.vectorize(lambda x: x.total_seconds())(diff))
		res_index = int(np.argmin(out))
		res_datetime = self.asDatetime(res_index)

		# res_datetime must be a dt.datetime since res_index is an int, so cast
		res_datetime = cast("dt.datetime", res_datetime)

		return res_datetime, res_index

	def areTimesWithin(self, t_search:dt.datetime|np.ndarray[tuple[int], np.dtype[np.datetime64]])\
						-> np.ndarray[tuple[int],np.dtype[np.bool_]]:
		"""Find if the provided times are within the timespan.

		Args:
			t_search: times to check if within timespan
				If timezone naive, assumed to be in UTC
				If timezone aware, will be converted to UTCtime to find

		Returns:
			ndarray of bools, True if within timespan
		"""
		if isinstance(t_search, dt.datetime):
			if t_search.tzinfo is not None:
				t_search = t_search.astimezone(dt.timezone.utc)
			else:
				t_search = t_search.replace(tzinfo=dt.timezone.utc)

		elif isinstance(t_search, np.ndarray):
			if t_search[0].tzinfo is not None:
				t_search = np.vectorize(lambda x:x.astimezone(dt.timezone.utc))(t_search)
			else:
				t_search = np.vectorize(lambda x:x.replace(tzinfo=dt.timezone.utc))(t_search)

		else:
			raise TypeError('t_search is of an unrecognised type,'
							' should be a datetime object or ndarray')

		return np.logical_and(t_search>np.asarray(self.start), t_search<np.asarray(self.end))

	def getFractionalIndices(self, t_search:dt.datetime|np.ndarray[tuple[int],
											np.dtype[np.datetime64]])\
							-> np.ndarray[tuple[int],np.dtype[np.float64]]:
		"""Find the fractional indices of timespan at wich t_search should be inserted.

		Find the indices in the original timespan at which each value of t_search should be
		inserted to maintain the sorted order.
		The integer part of the index indicates the value immediately prior to the value in
		t_search, while the fractional part represents the point between the two adjacent indices
		in the timespan at which the t_search value falls.
		For example (using integers rather than datetime objects:
			timespan = [0, 1, 5, 6, 10]
			t_search = [0.5, 2, 3, 8]
			timespan.getFractionalIndices(t_search) = [0.5, 1.25, 1.5, 3.5]
		All values of t_search must be within the timespan, otherwise the output is undefined.

		Args:
			t_search: times to locate within timespan
				If timezone naive, assumed to be in UTC
				If timezone aware, will be converted to UTCtime to find

		Returns:
			ndarray of fractional indices
		"""
		if isinstance(t_search, dt.datetime):
			if t_search.tzinfo is not None:
				t_search = t_search.astimezone(dt.timezone.utc)
			else:
				t_search = t_search.replace(tzinfo=dt.timezone.utc)
			t_search = np.asarray(t_search)
		elif isinstance(t_search, np.ndarray):
			if t_search[0].tzinfo is not None:
				t_search = np.vectorize(lambda x:x.astimezone(dt.timezone.utc))(t_search)
			else:
				t_search = np.vectorize(lambda x:x.replace(tzinfo=dt.timezone.utc))(t_search)
		else:
			raise TypeError('t_search is of an unrecognised type, '
							'should be a datetime object or ndarray')

		post_idxs = np.searchsorted(self._timearr, t_search) 		# type: ignore [arg-type]
		pre_idxs = post_idxs-1
		return (t_search - self._timearr[pre_idxs])\
				/(self._timearr[post_idxs]-self._timearr[pre_idxs]) + pre_idxs


	def _parseTimeperiod(self, t0:dt.datetime, timeperiod:str) -> dt.datetime:
		last_idx = 0
		for idx, letter in enumerate(timeperiod):
			last_idx = idx
			if not (letter.isdigit() or letter == '.'):
				break

		val = float(timeperiod[0:last_idx])
		unit = timeperiod[last_idx:]
		values = np.zeros(6)
		if unit == 'd':
			values[5] = val
		elif unit == 'H':
			values[4] = val
		elif unit == 'M':
			values[3] = val
		elif unit == 'S':
			values[2] = val
		elif unit == 'mS':
			values[0] = 1000 * val
		elif unit == 'uS':
			values[0] = val
		else:
			logger.error("Invalid timeperiod unit: Valid units are (y)ears, (m)onths, (W)eeks,"
						" (d)ays, (H)ours, (M)inutes, (S)econds, (mS) milliseconds,"
						" (uS) microseconds")
			raise ValueError(f"Invalid timeperiod unit:{unit}")

		return t0 + relativedelta(days=values[5], hours=values[4], minutes=values[3],
								seconds=values[2], microseconds=values[0])

	def _parseTimestep(self, timestep:str) -> dt.timedelta:
		last_idx = 0
		for idx, letter in enumerate(timestep):
			last_idx = idx
			if not (letter.isdigit() or letter == '.'):
				break

		val = float(timestep[0:last_idx])
		unit = timestep[last_idx:]
		values = np.zeros(6)
		if unit == 'd':
			values[5] = val
		elif unit == 'H':
			values[4] = val
		elif unit == 'M':
			values[3] = val
		elif unit == 'S':
			values[2] = val
		elif unit == 'mS':
			values[1] = val
		elif unit == 'uS':
			values[0] = val
		else:
			logger.error("Invalid timestep unit: Valid units are (d)ays, (H)ours," \
						" (M)inutes, (S)econds, (mS) milliseconds, (uS) microseconds")
			raise ValueError(f"Invalid timestep unit:{unit}")

		return dt.timedelta(days=values[5], seconds=values[2],
							microseconds=values[0], milliseconds=values[1],
							minutes=values[3], hours=values[4])

	def __len__(self) -> int:
		"""Returns the length of the TimeSpan."""
		return len(self._timearr)


	def cherryPickFromIndices(self, idxs:int|tuple|slice):
		"""Adjust TimeSpan to only contain the indices specified by idxs.

		Args:
			idxs: numpy style indexing of TimeSpan
		"""
		self._timearr = self._timearr[idxs]
		self.start = self._timearr[0]
		self.end = self._timearr[-1]
		self.time_step = None
		self.time_period = self.end - self.start

	@classmethod
	def fromDatetime(cls, dt_arr:np.ndarray[tuple[int], np.dtype[np.datetime64]],
							timezone:dt.timezone=dt.timezone.utc) -> 'TimeSpan':
		"""Create a TimeSpan from an array of datetime objects.

		Args:
			dt_arr: 1D array of datetime objects
			timezone: timezone to apply to each element of datetime array.
				Default: dt.timezone.utc
		"""
		for ii in range(len(dt_arr)):
			dt_arr[ii] = dt_arr[ii].replace(tzinfo=timezone)
		start = dt_arr[0]
		end = dt_arr[-1]
		tperiod = (end-start).total_seconds()
		tstep = tperiod/len(dt_arr)
		t = cls(start,f'{tstep}S',f'{tperiod}S')
		t._timearr = dt_arr.copy()
		t.start = start
		t.end = end
		t.time_step = None
		t.time_period = t.end-t.start

		return t



import configparser
from importlib import metadata, resources
import os
import pathlib

import platformdirs

from spherapy.util import credentials


def _creatPackagedTLEListing() -> None|dict[int,pathlib.Path]:
	packaged_tles = {}
	package_file_listing = metadata.files('spherapy')
	package_dir = resources.files('spherapy')
	if package_file_listing is not None and len(package_file_listing) > 0:
		for path in package_file_listing:
			if 'TLEs' in path.parts:
				try:
					tle_id = int(path.stem)
				except ValueError:
					print("Can't import packaged TLEs, skipping...") #noqa: T201
					return None
				traversable = package_dir.joinpath(f"{path.relative_to('spherapy')}")
				resource_path = pathlib.Path(f'{traversable}')
				packaged_tles[tle_id] = resource_path
	return packaged_tles

service_name = "spherapy"
service_author = "MSL"

try:
	__version__ = metadata.version(service_name)
except metadata.PackageNotFoundError:
	print(f'Version of {service_name} is uknown') #noqa: T201
	__version__ = "x.x.x"

config_dir_str = os.getenv('SPHERAPY_CONFIG_DIR')
if config_dir_str is None:
	use_config_file = False
	config_dir = pathlib.Path(platformdirs.user_config_dir(appname=service_name,
															version=__version__,
															ensure_exists=True))
else:
	use_config_file = True
	config_dir = pathlib.Path(config_dir_str)

# will load a config even if not using
config = configparser.ConfigParser()
read_files = config.read(config_dir.joinpath('spherapy.conf'))

if use_config_file and len(read_files) == 0:
	raise configparser.Error(f"Couldn't find spherapy.conf at {config_dir}")

# Locate dir to store TLE files
if not use_config_file:
	# use default user directories
	_data_dir = pathlib.Path(platformdirs.user_data_dir(appname=service_name, ensure_exists=True))
	tle_dir = _data_dir.joinpath('TLEs')
else:
	# use the config file
	_TLE_dir_str = config['paths'].get('TLE_path')

	if _TLE_dir_str is None:
		raise ValueError("Config file has not TLE_path")

	_temp_dir = pathlib.Path(_TLE_dir_str)
	if _TLE_dir_str == '':
		# TLE dir not set in config file
		_data_dir = pathlib.Path(platformdirs.user_data_dir(appname=service_name,
															ensure_exists=True))
		tle_dir = _data_dir.joinpath('TLEs')
	elif _temp_dir.is_absolute():
		tle_dir = _temp_dir
	else: 						# noqa: PLR5501
		# _temp_dir is relative
		if config_dir.is_dir():
			tle_dir = config_dir.joinpath(_TLE_dir_str)
		else:
			tle_dir = config_dir.parent.joinpath(_TLE_dir_str)

tle_dir.mkdir(parents=True, exist_ok=True)

# Load credentials

spacetrack_credentials:dict[str,str|None] = {'user':None, 'passwd':None}
# load spacetrack_credentials from the relevant source
if not use_config_file:
	credentials._reloadCredentials() 			# noqa: SLF001
else:
	credentials._reloadCredentials(config) 		# noqa: SLF001


packaged_TLEs = _creatPackagedTLEListing() 	#noqa: N816
